//bin.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;

static int bin_count = 0;

BIN::BIN()
{
	first_poly = last_poly = NULL;
	first_vert = last_vert = NULL;
	next = prev = NULL;
	marked = false;
	cache_next = cache_prev = NULL;
}
BIN::~BIN()
{
	//Clear bin_src for each poly in here
	POLY *poly = first_poly;
	while (poly)
	{
		poly->bin_src = NULL;
		poly = poly->bin_next;
	}
}

//Add a poly to bin
void BIN::add_poly(POLY *poly)
{
	if (!last_poly)	{	poly->bin_prev = NULL;		first_poly = poly;			}
	else			{	poly->bin_prev = last_poly;	last_poly->bin_next = poly;	}
	last_poly = poly;	poly->bin_next = NULL;		poly->bin_src = this;
}

//Not used!  May cause serious issues, so ignore for now
/*
void BIN::add_vert(VERT *vert)
{
	if (!last_vert)	{	vert->bin_prev = NULL;		first_vert = vert;			}
	else			{	vert->bin_prev = last_vert;	last_vert->bin_next = vert;	}
	last_vert = vert;	vert->bin_next = NULL;		vert->bin_src = this;
}
*/

//Remove a poly from the given bin, check if bin needs to be removed
void BIN::drop_poly(POLY *poly)
{
//	err_log << "DROPPING POLYGON" << endl;
	POLY *current = first_poly;
	bool found = false;
	while (current)
	{
		if (current==poly)
		{
			if (current->bin_prev)	current->bin_prev->bin_next = current->bin_next;
			else					first_poly = current->bin_next;
			if (current->bin_next)	current->bin_next->bin_prev = current->bin_prev;
			else					last_poly = current->bin_prev;
			current->bin_src = NULL;
			found = true;
			break;
		}
		current = current->bin_next;
	}
//	err_log << "                  FOUND: " << found;
	check_self();
}
/*
void BIN::drop_vert(VERT *vert)
{
	VERT *current = first_vert;
	while (current)
	{
		if (current==vert)
		{
			if (current->bin_prev)	current->bin_prev->bin_next = current->bin_next;
			else					first_vert = current->bin_next;
			if (current->bin_next)	current->bin_next->bin_prev = current->bin_prev;
			else					last_vert = current->bin_prev;
			current->bin_src = NULL;
			break;
		}
		current = current->bin_next;
	}
	check_self();
}
*/
//check if we need to get rid of the bin
void BIN::check_self()
{
	if (first_poly==NULL && first_vert==NULL)
	{
		frame.bins->remove_bin(this);
	}
}

//Clear the given bin
void BIN::clear()
{
	//Technically, we COULD go through and kill all those links.
	// HOWEVER - if we keep the links clean as we add/remove them,
	// then IT DOESN'T MATTER what they are left at.

	POLY *poly = first_poly;
	while(poly)
	{
		poly->bin_src = NULL;
		poly = poly->bin_next;
	}

	first_poly = last_poly = NULL;
	first_vert = last_vert = NULL;
}

//*****************************************************************************************
//*****************************************************************************************
//*****************************************************************************************
//*****************************************************************************************

//Create a list of the given dimensions -- determines bin size
// As you can see, the sizes can be non-uniform!  However, I do not
// take advantage of this, and generally just create uniform cubes
BIN_LIST::BIN_LIST(double x, double y, double z)
{
	first = NULL;
	last = NULL;
	wx = x;
	wy = y;
	wz = z;
	cache_first = cache_last = NULL;
	cache_max = 40;
	cache_count = 0;

	miss_cache_index = 0;
	miss_cache_max = 40;
	miss_cache = new BIN[miss_cache_max];
	miss_cache_count = 0;
}

BIN_LIST::~BIN_LIST()
{
	clear();

	//Miss cache should be only new memory
	delete [] miss_cache;
}

void BIN_LIST::clear_cache()
{
	cache_first = cache_last = NULL;
	cache_count = 0;
	miss_cache_count = 0;
	miss_cache_index = 0;
}

//Clear all bins, and delete all bins
void BIN_LIST::clear()
{
	BIN *current = first;
	BIN *remove;
	while (current)
	{
		remove = current;
		current = current->next;
		remove->clear();
		delete remove;
	}
	first = last = NULL;
}

//If bin is deleted, remove its entry from its neighbor's neighbor lists
void BIN::clear_from_neighborhood()
{
	//Take out neighbor entries from neighbors
	// n represents the neighbor of itself
	// N represents ITSELF as seen from neighbor n
	int a,b,c,n=0,N=26;
	for (a=-1;a<=1;a++)
		for (b=-1;b<=1;b++)
			for (c=-1;c<=1;c++)
			{
				if (neighbors[n])
					neighbors[n]->accept_neighbor(NULL,N);
				n++;
				N--;	//avoiding (26-n) 27 times
			}
}

//Remove the given bin, from main, cache, etc.  GIVE IT THE BOOT
void BIN_LIST::remove_bin(BIN *bin)
{
//	err_log << "BIN REMOVED!!!!!!!!!!!!!!! "  << bin << endl;
	remove_from_cache(bin);
	bin->clear_from_neighborhood();

	if (bin->prev) bin->prev->next = bin->next;
	else {first = bin->next; first->prev = NULL;}
	if (bin->next) bin->next->prev = bin->prev;
	else {last = bin->prev; last->next = NULL;}

	delete bin;
}

void BIN_LIST::remove_from_cache(BIN *bin)
{
	if (bin!=cache_last && bin->cache_next)
		bin->cache_next->cache_prev = bin->cache_prev;
	else
	{
		cache_last = bin->cache_prev;
		if (cache_last)
			cache_last->cache_next = NULL;
	}


	if (bin!=cache_first && bin->cache_prev)
		bin->cache_prev->cache_next = bin->cache_next;
	else
	{
		cache_first = bin->cache_next;
		if (cache_first)
			cache_first->cache_prev = NULL;
	}


	bin->cache_next = bin->cache_prev = NULL;
}

//And things start to get messy.
//Side-effects:
//		p now contains new bin position
//		(poly is dropped from source bin if it existed) <= not anymore
bool BIN_LIST::new_bin_needed(POLY *poly, int *p)
{
	double *pos = poly->vert[0]->pos;	//FIRST vert of poly is key

	//Figure out the indices for the bin from the given position
	get_indices(pos,p);

	//Obviously it's needed if bin_src does not exist
	if (!poly->bin_src)
		return true;

	//If target position is equal to current source's position, not needed
	if (poly->bin_src->i == p[0] && poly->bin_src->j == p[1] && poly->bin_src->k == p[2])
		return false;

	//If we reach here, then bin_src exists but differs from the target indices
	return true;
}

//Check to see if the bin is in the neighbors
bool BIN::poly_in_neighbors(POLY *poly, int *p)
{
	BIN *nbin;
	for (int n=0; n<27; n++)
	{
		nbin = neighbors[n];

		//Obviously, #13 is either ITSELF or NULL (in other words, it's undefined
		// and we should not expect it to be anything so ignore it)
		if (n==13 || !nbin) continue;

		if (p[0]==nbin->i && p[1]==nbin->j && p[2]==nbin->k)
		{
			poly->bin_src->drop_poly(poly);
			nbin->add_poly(poly);
			return true;
		}
	}
	return false;
}

//We know that poly is not in other bin and needs new bin
void BIN_LIST::force_add_poly(int *p, POLY *poly)
{
	BIN *bin;
	bool from_neighbors = false;

	if (poly->bin_src)
	{
		//If bin_src exists, we want to see if the new target position is found
		// within its neighbors.  Otherwise, we just drop the poly.
		//This is not necessary, but saves time -- we may as well try to
		// exploit the spatial cohesion if we have a lingering bin link

		from_neighbors = poly->bin_src->poly_in_neighbors(poly,p);
		if (from_neighbors)		return;
		else					poly->bin_src->drop_poly(poly);

		//If we didn't find the target bin there, we just drop the poly from
		// that lingering bin link, to be sure
	}

	//ELSE -- at this point, the poly IS NOT LINKED from any bin
	bin = get_binv(p);	//takes a while - set as last resort
	if (!bin) bin = create_bin(p[0],p[1],p[2]);	//If that failed, create new bin!
	bin->add_poly(poly);	//Either way, add poly to this bin
}

//Don't know anything about poly yet
void BIN_LIST::add_poly(POLY *poly)
{
//	err_log << "ADDING A POLY TO BIN_LIST: " << poly << ", " << poly->bin_src << endl;

	int p[3];	//p[] is used for indices for this session
	if (new_bin_needed(poly,p))	force_add_poly(p,poly);
	else						return;
}
/*
void BIN_LIST::add_vert(VERT *vert)
{
	BIN *bin;

	if (vert->bin_src)
		vert->bin_src->drop_vert(vert);

	float *pos = vert->pos;
	bin = get_binfv(pos);
	if (!bin) bin = create_bin(pos[0],pos[1],pos[2]);
	bin->add_vert(vert);

}
*/

//Create a bin with the given indices
BIN *BIN_LIST::create_bin(int i, int j, int k)
{
//	err_log << "Creating new bin..." << endl;
	bin_count++;

	BIN *bin = new BIN;
//	bin->i = x / wx;
//	bin->j = y / wy;
//	bin->k = z / wz;
//	if (x<0) bin->i--;	if (y<0) bin->j--;	if (z<0) bin->k--;

	bin->i = i;
	bin->j = j;
	bin->k = k;

	bin->x = bin->i * wx;
	bin->y = bin->j * wy;
	bin->z = bin->k * wz;

	bin->wx = wx;
	bin->wy = wy;
	bin->wz = wz;

//	err_log << "BINS: " << bin_count << endl;
//	err_log << bin_count << " Creating new bin: " << x << ", " << y << ", " << z << "  => " << bin->i << ", " << bin->j << ", " << bin->k << endl;

	//FIND NEIGHBORING BINS, ADD TO LIST
	int p[3];
	int n = 0;
	BIN *nbin;
	int a,b,c;
	for (a=-1;a<=1;a++)
		for (b=-1;b<=1;b++)
			for (c=-1;c<=1;c++)
			{
//				err_log << "ABC: " << a << ", " << b << ", " << c << endl;
//				a = 1;b=0;c=0;
				if (a==0&&b==0&&c==0)
				{
					bin->accept_neighbor(NULL,n);
				}
				else
				{
					p[0] = i+a; p[1] = j+b; p[2] = k+c;
//					err_log << "GET BIN!" << endl;
					nbin = get_binv(p);
//					err_log << "ACCEPT NEIGHBOR: " << nbin << endl;
					bin->accept_neighbor(nbin,n);
					if (nbin)
					{
//						err_log << "BIN: " << i << ", " << j << ", " << k << endl;
//						err_log << " Got neighbor: " << a << ", " << b << ", " << c << endl;

						//We want to be a neighbor of nbin!
						//Find inverse of n, and set that of nbin
						//INVERSE = 26 - n
						nbin->accept_neighbor(bin,26-n);
					}
				}
				n++;
			}


//	err_log << "            Adding this bin!" << endl;
	add_bin(bin);

//	err_log << "Bin CREATED" << endl;
	return bin;
}

//The neighboring bin will accept the given bin in the given slot
void BIN::accept_neighbor(BIN *neighbor, int n)
{
	neighbors[n] = neighbor;
}

//Add a given bin to the main list
void BIN_LIST::add_bin(BIN *bin)
{
	if (!last)
	{
		bin->prev = NULL;
		first = bin;
	}
	else
	{
		bin->prev = last;
		last->next = bin;
	}
	last = bin;
	bin->next = NULL;
}

//convert an absolute position to integer indices that uniquely identify a bin
void BIN_LIST::get_indices(double *pos, int *p)
{
	p[0] = pos[0] / wx;
	p[1] = pos[1] / wy;
	p[2] = pos[2] / wz;
	if (pos[0]<0) p[0]--;
	if (pos[1]<0) p[1]--;
	if (pos[2]<0) p[2]--;
}

//A few different ways to find bins
BIN *BIN_LIST::get_binfv(double *pos)
{
	int p[3];
	get_indices(pos,p);
	return get_binv(p);
}
BIN *BIN_LIST::get_binf(double x, double y, double z)
{
	int p[3];
	double pos[3];
	pos[0] = x;
	pos[1] = y;
	pos[2] = z;
	get_indices(pos,p);
	return get_binv(p);
}



//Check if bin is in miss cache
BIN *BIN_LIST::check_miss_cache(int i, int j, int k)
{
	//For consistency, this will return an address.  This is where
	// strangeness occurs -- if a valid address is returned, then we know
	// that the target is IN THE MISS CACHE, and therefore DOES NOT EXIST.
	//If NULL is returned, then it is not in the miss cache, but this DOES
	// NOT imply that it does not exist!  We will just need to check the
	// cache and the main list next.

	BIN dummy;	//Just used as a dummy, as would be expected.  Value is NOT USED
	for (int q=0; q<miss_cache_count; q++)
	{
		int Q = q + miss_cache_index;
		while (Q >= miss_cache_max) Q-=miss_cache_max;
		if (i==miss_cache[Q].i && j==miss_cache[Q].j && k==miss_cache[Q].k)
			return &dummy;	//Compiler DOES NOT LIKE THIS.  TOO BAD
	}
	return NULL;
}
//Find target from cache
BIN *BIN_LIST::check_cache(int i, int j, int k)
{
	BIN *current;
	current = cache_first;
	int n=0;
	while (current && n<cache_count)
	{
		if (current->i == i && current->j == j && current->k == k)
		{ frame.cache_hits++; return current; }
		current = current->cache_next;
		n++;
	}
	frame.cache_misses++;
	return NULL;
}
//Find target from main list (slowest!)
BIN *BIN_LIST::check_main(int i, int j, int k)
{
	BIN *current;
	current = first;
	while (current)
	{
		if (current->i == i && current->j == j && current->k == k)
		{ augment_cache(current); return current; }
		current = current->next;
	}
	return NULL;
}


//And the next three are mirrors of the above three, but with a pointer
// instead of three parameters.  In short, no different
BIN *BIN_LIST::check_miss_cache(int *p)
{
	BIN dummy;
	for (int q=0; q<miss_cache_count; q++)
	{
		int Q = q + miss_cache_index;
		while (Q >= miss_cache_max) Q-=miss_cache_max;
		if (p[0]==miss_cache[Q].i && p[1]==miss_cache[Q].j && p[2]==miss_cache[Q].k)
			return &dummy;
	}
	return NULL;
}
BIN *BIN_LIST::check_cache(int *p)
{
	BIN *current;
	current = cache_first;
	int n=0;
	//Potential issue!  Why do I need to check cache_count?  Something bad is happening.
	while (current && n<cache_count)
	{
		if (current->i == p[0] && current->j == p[1] && current->k == p[2])
		{ frame.cache_hits++; return current; }
		current = current->cache_next;
		n++;
	}
	frame.cache_misses++;
	return NULL;
}
BIN *BIN_LIST::check_main(int *p)
{
	BIN *current;
	current = first;
	while (current)
	{
		if (current->i == p[0] && current->j == p[1] && current->k == p[2])
		{ augment_cache(current); return current; }
		current = current->next;
	}
	return NULL;
}


//Check main list, but DO NOT AUGMENT CACHE with results!
// This is used by searching functions that do not reference cache!
// Otherwise, it would fragment the cache horrifically
BIN *BIN_LIST::check_mainv(int *p)
{
	BIN *current;
	current = first;
	while (current)
	{
		if (current->i == p[0] && current->j == p[1] && current->k == p[2])
			return current;
		current = current->next;
	}
	return NULL;
}





//Find a bin from the given indices
// Unlike get_binv - this one uses cache.
BIN *BIN_LIST::get_bin(int *p)
{
	BIN *r = NULL;
	if (check_miss_cache(p)) return NULL;

	r = check_cache(p);
	if (r) return r;
	
	r = check_main(p);
	if (r) return r;
	
	augment_miss_cache(p[0],p[1],p[2]);
	frame.cache_misses--;
	
	return NULL;
}
//Mirror of above, with different input type
BIN *BIN_LIST::get_bin(int i, int j, int k)
{
	BIN *r = NULL;
	if (check_miss_cache(i,j,k)) return NULL;

	r = check_cache(i,j,k);
	if (r) return r;

	r = check_main(i,j,k);
	if (r) return r;

	augment_miss_cache(i,j,k);
	frame.cache_misses--;

	return NULL; 
}

//ONLY checks main list, without any cache checking or augmenting
BIN *BIN_LIST::get_binv(int *p)
{
	BIN *r;

	r = check_mainv(p);
	if (r) return r;

	return NULL;
}

//Draw bins, for testing.  This could be made more efficient, but
// the effort is not really worth it -- it is purely debugging
void BIN_LIST::draw()
{
	int i;
	BIN *current;
	int target = 59;

	if (frame.config.get_bool(VAR_DBINS))
	{
		glBegin(GL_LINES);
		i = 0;
		current = first;
		while (current)
		{
			current->draw();
			i++;
			current = current->next;
		}
		glEnd();
	}
}
void BIN::draw()
{
//	if (!marked) return;

	double x0 = (x);// - wx/2.0f;
	double y0 = (y);// - wy/2.0f;
	double z0 = (z);// - wz/2.0f;

	double x1 = x0+wx;
	double y1 = y0+wy;
	double z1 = z0+wz;

	float pad = .0001f;

	x0 += pad;
	y0 += pad;
	z0 += pad;
	x1 -= pad;
	y1 -= pad;
	z1 -= pad;

	if (!marked)	glColor4f(1,0,0,1);
	else			glColor4f(0,1,0,1);

	if (!first_poly) glColor4f(1,.5f,0,1);


	glVertex3f(x0,	y0,	z0);
	glVertex3f(x1,	y0,	z0);
	glVertex3f(x1,	y0,	z0);
	glVertex3f(x1,	y1,	z0);
	glVertex3f(x1,	y1,	z0);
	glVertex3f(x0,	y1,	z0);
	glVertex3f(x0,	y1,	z0);
	glVertex3f(x0,	y0,	z0);

	glVertex3f(x0,	y0,	z1);
	glVertex3f(x1,	y0,	z1);
	glVertex3f(x1,	y0,	z1);
	glVertex3f(x1,	y1,	z1);
	glVertex3f(x1,	y1,	z1);
	glVertex3f(x0,	y1,	z1);
	glVertex3f(x0,	y1,	z1);
	glVertex3f(x0,	y0,	z1);

	glVertex3f(x0,	y0,	z0);
	glVertex3f(x0,	y0,	z1);
	glVertex3f(x1,	y0,	z0);
	glVertex3f(x1,	y0,	z1);
	glVertex3f(x0,	y1,	z0);
	glVertex3f(x0,	y1,	z1);
	glVertex3f(x1,	y1,	z0);
	glVertex3f(x1,	y1,	z1);

}






//The masterful collision detection stuff, for a given BIN
void BIN::collide_proj2poly(PHYS_OBJ *proj)
{
	double *origin;
	double *vector;
	POLY *poly = first_poly;
	int  r;

	//Grab a pointer to the given projectile's last_collision structure
	CLOSEST_COLLISION *closest = &proj->last_poly_collision;

	//If a bin is used at all, mark it -- debugging
	mark();

	//We COULD set this in BIN_LIST, but then we'd have to send it.  Is it worth the
	// extra parameters to save, oh, 9 assignments or so (assuming that many bins in neighborhood)?
	//ON SECOND THOUGHT - do we need this?  We might only call proj->pos a couple times as it is
	origin = proj->pos;
	vector = proj->vel;

	//These will be passed to collision detection functions
	double t, u, v;
	t = u = v = 0;

	while (poly)
	{
		//check collision!

		r = intersect_triangle(origin, vector, poly->vert[0]->pos, poly->vert[1]->pos, poly->vert[2]->pos, &t, &u, &v);

		if (r==1)
		{
			// <= 1 implies that vector of movement INCLUDES final point of o+v
			// > 0 implies that vector of movement EXCLUDES starting pos o (thus,
			//      if we are starting ON a polygon, but moving AWAY, it will not
			//      register a collision.

			if (t<=1 && t>0)
			{
//				err_log << "A collision is found!" << endl;
//				err_log << "Result (r,t,u,v): " << r << ", " << t << ", " << u << ", " << v << endl;

//				err_log << "Poly ID: " << poly->id << endl;

//				frame.pvector("POLY: ",poly->vert[0]->pos);
//				frame.pvector("PROJ: ",proj->pos);
//				frame.pvector("VEL: ",proj->vel);

				if (!closest->poly)
				{
//					err_log << "NEW: " << poly << endl;
//					err_log << "Setting closest for first time: " << *closest_t << " to " << t << endl;
					closest->t = t;
					closest->poly = poly;
					closest->u = u;
					closest->v = v;
				}
				else
				{
					if (closest->t > t)
					{
//						err_log << "CLOSER" << endl;
//						err_log << "We have a closer collision! " << *closest_t << " to " << t << endl;
						closest->t = t;
						closest->poly = poly;
						closest->u = u;
						closest->v = v;
					}
				}

//				err_log << "GAR! " << closest << ", " << closest->poly << endl;
			}
		}

		//if collision is closer, or closest==null, assign closest to this poly
		poly = poly->bin_next;
	}
}

//Reset collision so that any later ones overwrite it
void CLOSEST_COLLISION::reset()
{
	poly = NULL;
	t = 4000000000;
	speed = 1;
}

//Overall bin finding for collisions.  Comment blocks are
// old methods I used for collecting bins, and are slower, but
// may still be potentially useful so I'm saving them
void BIN_LIST::collide_proj2poly(PHYS_OBJ *proj)
{
	BIN *bin = NULL;
	proj->last_poly_collision.reset();

	int i,j,k;
	int p[3];

	double pos2[3];
	int p2[3];

	//This still is just grabbing the bin that has ray's origin
	// This is not so great, but particles don't generally go fast enough for that to matter
	// (they'd have to move through an entire bin width in one step, and that
	// is TOO FAST in this model anyway).

	//Imagine that this block is PER BIN.  Thus, inside it will remain solid,
	// while outside it, we may just expand the number of bins visited.

	i=j=k=0;
	get_indices(proj->pos, p);


/*
	for (i=-1;i<=1;i++)
		for (j=-1;j<=1;j++)
			for (k=-1;k<=1;k++)
			{
				bin = get_bin(p[0]+i,p[1]+j,p[2]+k);

				if (bin)
				{
					closest = bin->collide_proj2poly(proj, closest, &closest_t);
				}
			}
*/



	//If in same bin still, don't bother searching for new
	BIN *self_bin = proj->bin_src;
	if (self_bin && self_bin->i==p[0] && self_bin->j==p[1] && self_bin->k==p[2])
			bin = self_bin;
	else	bin = get_bin(p);


	//Check neighbors if bin exists
	if (!bin)
	{
		//If source bin doesn't exist, check destination bin

		pos2[0] = proj->pos[0] + proj->vel[0];
		pos2[1] = proj->pos[1] + proj->vel[1];
		pos2[2] = proj->pos[2] + proj->vel[2];
		get_indices(pos2,p2);

		if (p2[0]!=p[0] || p2[1]!=p[1] || p2[2]!=p[2])
			bin = get_bin(p2);
	}

	if (bin)
	{
		//This is not remarkably faster, but has the potential
		// to sometimes be better in the end.

		//IDEA: If collision found in self bin
		// then IGNORE NEIGHBORS.  Is this ok?  LET'S HOPE SO
		bin->collide_proj2poly(proj);

		BIN **N;

		//I theorize that IT IS OK.  We will leave
		// it as such for now, but if I find a reason
		// to change back, just remove the line below
		if (!proj->last_poly_collision.poly)
		{
			N = bin->neighbors;
			for (i=0;i<27;i++)
			{
				if (N[0])
				{
					N[0]->collide_proj2poly(proj);
				}
				N++;
			}
		}
	}

/*

	else
	{
		for (i=-1;i<=1;i++)
			for (j=-1;j<=1;j++)
				for (k=-1;k<=1;k++)
				{
					if (i==0&&j==0&&k==0) continue;


					P[0] = p[0]+i;P[1] = p[1]+j;P[2] = p[2]+k;
//					bin = get_bin(P[0],P[1],P[2]);
//					bin = get_bin(p[0]+i,p[1]+j,p[2]+k);
					bin = get_binv(P);

					if (bin)
					{
						bin->collide_proj2poly(proj);
					}
				}
	}
*/

	if (proj->last_poly_collision.poly)
	{
//		err_log << "Closest set, we have a collision!" << endl;
		proj->collide_poly();
	}

/*
	bin = get_binfv(proj->pos);
	if (bin)
	{
		bin->collide_proj2poly(proj, closest);
	}
*/

	//After we're done with this set, closest will be set to the closest
	// polygon collided.  If it's null, then no collision!  Whoah!

	//Side note: Collisions only count if they hit the front face -- otherwise
	// it just goes through.
}



void BIN_LIST::augment_cache(BIN *bin)
{
	int q=0;
	BIN *c = cache_first;

	//Make sure there are no repeats
	while (c)
	{
		q++;
		if (bin==c)
		{
			//Bin already in cache, ignore
			return;
		}
		c = c->cache_next;
	}


	//Add new entry
	if (!cache_last || !cache_first)
	{
		bin->cache_prev = NULL;
		cache_first = bin;
	}
	else
	{
		bin->cache_prev = cache_last;
		cache_last->cache_next = bin;
	}
	cache_last = bin;
	bin->cache_next = NULL;


	//If we have exceeded the cache count, knock out the oldest
	if (cache_count<cache_max)
	{
		cache_count++;
	}
	else
	{
//		err_log << "Moving first bin along." << endl;
		if (cache_first)
		{
			BIN *temp = cache_first;
			cache_first = cache_first->cache_next;
			temp->cache_next = NULL;
			temp->cache_prev = NULL;
		}
	}
}


//We found bin coords that have no bin -- add to miss cache
void BIN_LIST::augment_miss_cache(int I, int J, int K)
{
	int i = miss_cache_index + 1;
	if (i>=miss_cache_max) i -= miss_cache_max;

	if (miss_cache_count<miss_cache_max)
		miss_cache_count++;

	miss_cache[i].i = I;
	miss_cache[i].j = J;
	miss_cache[i].k = K;

	miss_cache_index++;
	if (miss_cache_index >= miss_cache_max)
		miss_cache_index = 0;

}







//Hopefully the comment below remains true:
//STILL under 1000 lines !



