//bin.h
/*
wherein the fate of the world is decided

This is my haphazard binning code specifically designed
for this project.  It is scary.  People have died within it.

The basic idea is that there is a BIN_LIST, which stores a pointer
to the FIRST and LAST bins.  Each BIN is a node in this list,
and they can be added or removed dynamically.  Each BIN has a
location, which is used as a unique identifier, and has a
FIRST and LAST pointer to its stored objects.

As for stored objects, it has code to store POLYs and VERTs,
but I have only implemented the use of POLYs.  Each POLY
has a bin_next and bin_prev which is used to create their
own linked list, which works because each POLY can only be
in a single BIN at a time, so each BIN has its own ad hoc
linked list.

The function add_poly() does two functions: If a poly is already
stored in a bin (referenced by bin_src), then it needs to be
removed from that bin, and then added to the new bin.  The
new bin is determined by the POLY's position, divided by
the bin dimensions.  If the potential new bin would have the
same position as the old bin, then nothing happens.  Otherwise,
the POLY is moved to the new BIN.  If a POLY is removed from a
BIN, and that BIN no longer contains any POLYs, then that BIN is
removed from the list and deleted.  If a POLY is added, and is
calling for a BIN position that does not exist, the BIN is created,
and added to the list, wherein the POLY is then added to that BIN.

When deciding if a BIN exists, or to just find it, all the BINs
have to be cycled through, comparing their positions to the
target position.  If there are many bins, this is atrocious, so
I made a caching system, such that there is a smaller set of
most-recently used bins that it can search first.  Since photons
move one at a time, this generally works well, as the photon
has to search the bins to find the nearest ones, and will only
reference a smaller handful at a time.  There is also a miss
cache, that signifies which positions WILL NOT contain a BIN,
so that a photon travelling through empty space will not constantly
be traversing the list.  As you might expect, these various caches
are updated when BINs are added or removed.

As if that wasn't enough for nightmares, I also have each BIN
keep track of its neighbors.  When a BIN is added, it figures
out the positions of its neighboring bins and tries to find
them from the main list (cache is not used as the search is 
very fragmented).  If found, it stores a link in an array of
length 27 (3x3x3), where the order determines which direction
the neighbor is in (to find the reciprocal neighbor, you'd
reference 26-i in the neighbor's neighbors).  When BINs are
removed, they check their neighbors, and remove any links to
themselves.

And the best news is that if I had made this hierarchical instead,
I would not have needed any of the caching systems and it would
have been significantly faster and simpler.  It is best
that you do not think too much about the bins for now.

*/

/*
Random side notes I have made:

Shall we assume that a triangle CANNOT EVER span
more than neighboring bins?  So, if it were only in bins for which
its vertices reside, it'd be fine?  This will ONLY work, also
if we read from neighboring bins when working, as it is possible at
corners to violate this.  BUT ONLY if triangles remain small compared
to bin size - otherwise they can envelop whole bins, in which
case, bin size would be pretty useless at that level anyway (less than
one triangle per bin?)

Also: bin:poly = 1:1, or else bin_next etc won't work right (we'll be
overwriting it).  We COULD use a separate element of bin_next[3] for
each vertex, but that'd be extra complications on traversal (such as,
which to use?  Does that mean we'd have to branch from each poly,
potentially?), so no, nevermind.  We'll just mark each poly from its
FIRST VERTEX (vert[0]), and hope their size remains small enough that
you'll encapsulate all polys touching the middle bin when browsing
ALL neighbors.  But we must keep bins to a size such as that already,
so we should be fine.

Unfortunately, bin size is relative to the composition of the model,
so it'll have to be fluid.  Maybe an analyzer on reading in the model
will figure out the average poly max length, and create the bin
size from that.
*/



#ifndef _BIN_
#define _BIN_


//When finding collisions, this keeps track of the closest one
// (which is all that will matter).
struct CLOSEST_COLLISION
{
	void reset();	//Clear before looking for collisions
	POLY *poly;		//Poly that caused collision
	double t;		//Percentage along ray where intersection occured
	double u;		//u,v triangle coordinates of collision
	double v;
	double speed;	//Speed at which collision occurred
};

class BIN
{
public:
	BIN();
	~BIN();

	void add_poly(POLY *poly);
	void add_vert(VERT *vert);
	void drop_poly(POLY *poly);
	void drop_vert(VERT *vert);

	void check_self();

	void clear();
	void clear_from_neighborhood();
	bool poly_in_neighbors(POLY *poly, int *p);

	void draw();

	void accept_neighbor(BIN *neighbor, int n);

	BIN *cache_next;
	BIN *cache_prev;

	BIN *next;
	BIN *prev;
	int i,j,k;
	double x,y,z;

	double wx, wy, wz;

	void collide_proj2poly(PHYS_OBJ *proj);

	void mark() {marked = true;}

	BIN *neighbors[27];

	POLY *first_poly;
	POLY *last_poly;
	VERT *first_vert;
	VERT *last_vert;

protected:
	bool marked;
};


class BIN_LIST
{
public:
	BIN_LIST(double x, double y, double z);
	~BIN_LIST();

	bool new_bin_needed(POLY *poly, int *p);
	void force_add_poly(int *p, POLY *poly);

	void add_poly(POLY *poly);
	void add_vert(VERT *vert);

	void remove_bin(BIN *bin);

	void clear();
	void clear_cache();

	void draw();

	void get_indices(double *pos, int *p);

	float get_wx() {return wx;}
	float get_wy() {return wy;}
	float get_wz() {return wz;}

	void collide_proj2poly(PHYS_OBJ *proj);

	void augment_cache(BIN *bin);
	void augment_miss_cache(int i, int j, int k);

protected:
	void add_bin(BIN *bin);
	BIN *create_bin(int i, int j, int k);

	BIN *get_binfv(double *pos);
	BIN *get_binf(double x, double y, double z);

	BIN *get_bin(int *p);
	BIN *get_bin(int i, int j, int k);
	BIN *get_binv(int *p);

	BIN *check_miss_cache(int i, int j, int k);
	BIN *check_cache(int i, int j, int k);
	BIN *check_main(int i, int j, int k);
	BIN *check_miss_cache(int *p);
	BIN *check_cache(int *p);
	BIN *check_main(int *p);

	BIN *check_mainv(int *p);

	void remove_from_cache(BIN *bin);

	//Caching stuff
	int cache_max;
	int cache_count;
	BIN *cache_first;
	BIN *cache_last;
	int miss_cache_index;
	BIN *miss_cache;
	int miss_cache_max;
	int miss_cache_count;


	BIN *first;
	BIN *last;

	//Bin dimensions
	double wx,wy,wz;
};




#endif


