//model.cpp

#include "main.h"

extern FRAME frame;
extern LOGGER err_log;
extern double rnd();

//Unique poly ids
static long last_poly_id = -1;

//Reading function
static int vertex_cb(p_ply_argument argument);
static int face_cb(p_ply_argument argument);
static int parameters_cb(p_ply_argument argument);

//These are used to determine the extents of all the vertices
static double X0 = 10000;
static double X1 = -10000;
static double Y0 = 10000;
static double Y1 = -10000;
static double Z0 = 10000;
static double Z1 = -10000;

//Used while reading in models
int current_pindex = 0;
int current_vindex = 0;
int tri_index = 0;
int total_vert_count;
int total_poly_count;
MODEL *current_model;
VERT *current_vert;
VERT *tri[4];
VERT **temp_vert_list;

//Model parameter defaults
float new_durability = .75f;
bool new_invert = true;
static int which_parameter = 0;

//Loads a PLY file into a model & initializes model
bool MODEL::load_ply(char *str, char *new_name)
{
	inverted = false;

	err_log << "Loading model [" << new_name << "]: " << str << endl;
	sprintf(name,"%s",new_name);
	long nvertices, ntriangles;
	ply = ply_open(str, NULL);
	if (!ply)
	{
		frame.error("Cannot open ply file %s",str);
		return false;
	}
	if (!ply_read_header(ply))
	{
		frame.error("Cannot read header for ply file %s",str);
		return false;
	}


	current_pindex = 0;
	current_vindex = 0;
	tri_index = 0;


	char tmp[512];
	ply_get_counts(ply, &total_vert_count, &total_poly_count, tmp);

	current_model = this;

	temp_vert_list = new VERT*[total_vert_count];

	nvertices =		ply_set_read_cb(ply, "vertex", "x", vertex_cb, NULL, 0);
					ply_set_read_cb(ply, "vertex", "y", vertex_cb, NULL, 0);
					ply_set_read_cb(ply, "vertex", "z", vertex_cb, NULL, 1);

	current_pindex = 0;
	ntriangles =	ply_set_read_cb(ply, "face", "vertex_indices", face_cb, NULL, 0);


	ply_set_read_cb(ply, "parameters", "durability", parameters_cb, NULL, 0);
	ply_set_read_cb(ply, "parameters", "invert", parameters_cb, NULL, 0);


	if (!ply_read(ply))
	{
		frame.error("Cannot read ply file %s",str);
		return false;
	}

	durability = new_durability;
	if (new_invert) do_invert();

	printf("Num vertices: %ld\nNum triangles: %ld\n", nvertices, ntriangles);
	ply_close(ply);
	init();

	delete [] temp_vert_list;

	double padxy = .001f;
	double padz = .03f;
	frame.config.set(VAR_SIM_X0,X0-padxy);
	frame.config.set(VAR_SIM_X1,X1+padxy);
	frame.config.set(VAR_SIM_Y0,Y0-padxy);
	frame.config.set(VAR_SIM_Y1,Y1+padxy);
	frame.config.set(VAR_SIM_Z0,Z1+padz);
	frame.config.set(VAR_SIM_Z1,Z1+padz);

	//If we are using the giant bin model, then size the bins!
	if (frame.config.get_bool(VAR_GIANT_BIN))
	{
		float pad = .01f;
		float w;
		float w0 = (X1-X0) + pad*2;
		float w1 = (Y1-Y0) + pad*2;
		float w2 = (Z1-Z0) + pad*2;
		w = w0;
		if (w1>w) w = w1;
		if (w2>w) w = w2;

		frame.bin_width = w;
	}

	return true;
}


//Read vertex from ply
static int vertex_cb(p_ply_argument argument)
{
	long eol;
	ply_get_argument_user_data(argument, NULL, &eol);

	if (current_vindex>=3)
		return 1;
	if (current_vindex==0)
	{
		current_vert = current_model->add_vert(current_pindex);
		temp_vert_list[current_pindex] = current_vert;
	}
	current_vert->pos[current_vindex] = ply_get_argument_value(argument);

	current_vindex++;
	if (eol)
	{
		current_vindex = 0;
		current_pindex++;
	}
	return 1;
}

//Read polygon from ply (we're assuming they're all triangles!)
static int face_cb(p_ply_argument argument)
{
	long length, value_index;
	POLY *new_poly;
	ply_get_argument_property(argument, NULL, &length, &value_index);

	if (tri_index==0)
	{
		if ((int)ply_get_argument_value(argument)!=3)
			frame.fail("Model contains polygons that are not triangles.  How quaint and rude.");
	}
	else
		tri[tri_index] = temp_vert_list[(int)ply_get_argument_value(argument)];

	tri_index++;
	switch (value_index)
	{
	default: break;
	case 2:
		{
			new_poly = current_model->add_poly(3, tri[1], tri[2], tri[3]);

			tri[1]->add_adj(new_poly);
			tri[2]->add_adj(new_poly);
			tri[3]->add_adj(new_poly);
			tri_index = 0;
			break;
		}
	}
	return 1;
}

//Get parameters
static int parameters_cb(p_ply_argument argument)
{
	int DURABILITY = 0;
	int INVERT = 1;

	if (which_parameter==DURABILITY)
		new_durability = ply_get_argument_value(argument);
	if (which_parameter==INVERT)
		new_invert = (ply_get_argument_value(argument) != 0);

	which_parameter++;

	return 1;
}

void MODEL::do_invert()
{
	POLY *poly = first_poly;
	VERT *v[3];

	while (poly)
	{
		v[0] = poly->vert[0];
		v[1] = poly->vert[1];
		v[2] = poly->vert[2];

		poly->vert[0] = v[2];
		poly->vert[1] = v[1];
		poly->vert[2] = v[0];

		poly = poly->next;
	}

	inverted = !inverted;
}


//********************************************************************************************
//************************************************************************************ CONSTRUCTORS
//********************************************************************************************

EDGE::EDGE()
{
	next = NULL;
}
VERT::VERT(double *p)
{
	zero_vars();
	pos[0] = p[0];
	pos[1] = p[1];
	pos[2] = p[2];
}
VERT::VERT()
{
	zero_vars();
}

void VERT::zero_vars()
{
	num_adj = 0;
	num_adj_p = 0;
	edge_explored = false;
	vert_explored = false;
	next = NULL;
	orphan = true;
	bin_next = bin_prev = NULL;
	bin_src = NULL;
	src_model = NULL;
	marked = false;
	color[0]=color[1]=color[2]=color[3] = 1;
	color[1]=0;
}
POLY::POLY(int new_vert_count)
{
	vert_count = new_vert_count;
	vert[0] = vert[1] = vert[2] = NULL;
	neighbor[0] = NULL;
	neighbor[1] = NULL;
	neighbor[2] = NULL;
	next = NULL;
	prev = NULL;
	bin_next = bin_prev = NULL;
	bin_src = NULL;
	marked = false;
	src_model = NULL;
	subdiv_needed = false;
	subdiv_performed = false;
	vert[0]=vert[1]=vert[2] = NULL;
	longest_edge = 0;
}
MODEL::MODEL()
{
	ready = false;
	first_poly = NULL;
	last_poly = NULL;
	poly_count = 0;
	vert_count = 0;
	edge_count = 0;
	inverted = false;
	last_edge = NULL;
	display_list[0] = display_list[1] = display_list[2] = display_list[3] = -1;
//	va_ready = false;
	sprintf(name,"[Untitled]");
	durability = 1;
}

void MODEL::destroy()
{
	if (!ready) return;

	POLY *rpoly, *poly = first_poly;
	VERT *rvert, *vert = first_vert;
	EDGE *redge, *edge = first_edge;

	while (poly) {rpoly = poly; poly = poly->next; delete rpoly;}
	while (vert) {rvert = vert; vert = vert->next; delete rvert;}
	while (edge) {redge = edge; edge = edge->next; delete redge;}

	err_log << "Model destroyed." << endl;
}



//********************************************************************************************
//************************************************************************************ INITIALIZATION
//********************************************************************************************

//Must be called once.  Initializes normals for vertices and polygons, but
// most importantly, creates edge list.
void MODEL::init()
{
	if (ready)
	{
		frame.error("Attempting to init() a model more than once.");
		return;
	}
	ready = true;


	err_log << " Parse edges:" << endl;
	parse_edges();

	err_log << " Parse polys:" << endl;
	parse_polys();

	recalculate();

	err_log << "*** Model initialized *** " << endl;
	err_log << "Vertices: " << vert_count << endl;
	err_log << "Polys:    " << poly_count << endl;
	err_log << "Edges:    " << edge_count << endl << endl;
}

//Add all polys to bins
void MODEL::init_bins()
{
	if (!frame.bins)
		return;

	//polys are RELOCATING, which is handled in
	// add_poly.  It's like magic, I know.

	err_log << "INITIALIZING BINS" << endl;
	POLY *current;

	frame.display.overlay->set_wait_reason("Ininitializing bins");

	int n = 1;
	current = first_poly;
	while (current)
	{
		if (n%1000==0 || n==poly_count)
			frame.display.overlay->post_progress((float)n/(float)poly_count);

		frame.bins->add_poly(current);

		current = current->next;
		n++;
	}

	err_log << "Bins Initialized for model" << endl;
}
//Initialize normals for polys
void MODEL::init_poly_normals()
{
//	frame.messages->add("INIT POLY NORMALS");
	POLY *current;
	current = first_poly;

	while (current)
	{
		current->init_normal(inverted);
		current = current->next;
	}
}
//Initialize normals for verts (called AFTER init_poly_normals())
void MODEL::init_vert_normals()
{
	VERT *current;
	current = first_vert;
	while (current)
	{
		current->init_normal();
		current = current->next;
	}
}
//Init the normal at a given vert
void VERT::init_normal()
{
	if (num_adj==0)
	{
		//If a vert has no neighbors, then it is not part of a poly!
		orphan = true;
		return;
	}
	orphan = false;
	normal[0] = 0;
	normal[1] = 0;
	normal[2] = 0;
	for (int j=0; j<num_adj; j++)
	{
		normal[0] += adj[j]->normal[0];
		normal[1] += adj[j]->normal[1];
		normal[2] += adj[j]->normal[2];
	}
	normal[0] /= num_adj;
	normal[1] /= num_adj;
	normal[2] /= num_adj;
	color[3] = 1;
	dnormal[0] = normal[0] * NORMAL_DISPLAY_SCALE;
	dnormal[1] = normal[1] * NORMAL_DISPLAY_SCALE;
	dnormal[2] = normal[2] * NORMAL_DISPLAY_SCALE;
	residue.fix();
}
//Calculate normal of poly
void POLY::init_normal(bool inverted)
{
	normal_face(normal,vert[0]->pos,vert[1]->pos,vert[2]->pos);
}
//Create edge list.
void MODEL::parse_edges()
{
	frame.display.overlay->show_calculating("Parse edges...");
	err_log << "PARSE EDGES" << endl;
	int j,k;
	VERT *p;
	POLY *a;
	VERT *v;

	EDGE *e;
	EDGE *re;

	//Kill all edges!
	if (first_edge)
	{
		err_log << "CLEAR EDGES FIRST" << endl;
		e = first_edge;
		while (e)
		{
			e->end[0]->vert_explored = false;
			e->end[1]->vert_explored = false;
			e->end[0]->edge_explored = false;
			e->end[1]->edge_explored = false;

			re = e;
			e = e->next;
			delete re;
		}
		first_edge = last_edge = NULL;
		edge_count = 0;
		err_log << "EDGES CLEARED" << endl;
	}

	p = first_vert;
	while (p)
	{
		p->vert_explored = true;

		for (j=0; j<p->num_adj; j++)
		{
			a = p->adj[j];	//Adjacent triangle

			for (k=0; k<a->vert_count; k++)
			{
				v = a->vert[k];	//Corner of adjacent triangle

				if (!v->vert_explored)
				{
					p->add_adj_p(v);
					v->vert_explored = true;
					if (!v->edge_explored)
					{
						add_edge(p,v);
					}
				}
			}
		}

		//Clear vert_explored in neighbors (only used within each vertex's own neighbor scope!)
		for (j=0; j<p->num_adj; j++)
		{
			a = p->adj[j];
			for (k=0; k<a->vert_count; k++)
				a->vert[k]->vert_explored = false;
		}
		p->edge_explored = true;
		p = p->next;
	}
}

//Discover the neighbors of each triangle
// ALSO, find longest edge
void MODEL::parse_polys()
{
	frame.display.overlay->set_wait_reason("Parsing polys");
	int i;
	POLY *poly = first_poly;
	POLY *n;


	int c = 1;
	while (poly)
	{
		if (c%1000==0 || c==poly_count)
			frame.display.overlay->post_progress((float)c/(float)poly_count);

		for (i=0; i<poly->vert[0]->num_adj; i++)
		{
			n = poly->vert[0]->adj[i];
			if (n==poly) continue;
			if ((poly->vert[0] == n->vert[0] || poly->vert[0] == n->vert[1] || poly->vert[0] == n->vert[2]) &&
				(poly->vert[1] == n->vert[0] || poly->vert[1] == n->vert[1] || poly->vert[1] == n->vert[2]))
				poly->neighbor[0] = n;

			if ((poly->vert[0] == n->vert[0] || poly->vert[0] == n->vert[1] || poly->vert[0] == n->vert[2]) &&
				(poly->vert[2] == n->vert[0] || poly->vert[2] == n->vert[1] || poly->vert[2] == n->vert[2]))
				poly->neighbor[2] = n;
		}
		for (i=0; i<poly->vert[1]->num_adj; i++)
		{
			n = poly->vert[1]->adj[i];
			if (n==poly) continue;
			if ((poly->vert[1] == n->vert[0] || poly->vert[1] == n->vert[1] || poly->vert[1] == n->vert[2]) &&
				(poly->vert[2] == n->vert[0] || poly->vert[2] == n->vert[1] || poly->vert[2] == n->vert[2]))
				poly->neighbor[1] = n;
		}

		c++;
		poly = poly->next;
	}
}

void POLY::set_longest_edge()
{
	//Find longest edge
	double longest = 0;
	double *tmp[3];
	double d;
	tmp[0] = vert[0]->pos;
	tmp[1] = vert[1]->pos;
	tmp[2] = vert[2]->pos;
	d = pow(vert[0]->pos[0]-vert[1]->pos[0],2) + 
		pow(vert[0]->pos[1]-vert[1]->pos[1],2) + 
		pow(vert[0]->pos[2]-vert[1]->pos[2],2);
	longest = d;

	d = pow(vert[1]->pos[0]-vert[2]->pos[0],2) + 
		pow(vert[1]->pos[1]-vert[2]->pos[1],2) + 
		pow(vert[1]->pos[2]-vert[2]->pos[2],2);
	if (d>longest) longest = d;

	d = pow(vert[0]->pos[0]-vert[2]->pos[0],2) + 
		pow(vert[0]->pos[1]-vert[2]->pos[1],2) + 
		pow(vert[0]->pos[2]-vert[2]->pos[2],2);
	if (d>longest) longest = d;

	longest_edge = sqrt(longest);

	if (longest_edge > frame.absolute_longest_edge)
	{
		frame.absolute_longest_edge = longest_edge;
	}
}

bool POLY::is_neighbor(POLY *n)
{
	int matches = 0;
	if (n->vert[0] == vert[0] || n->vert[0] == vert[1] || n->vert[0] == vert[2]) matches++;
	if (n->vert[1] == vert[0] || n->vert[1] == vert[1] || n->vert[1] == vert[2]) matches++;
	if (n->vert[2] == vert[0] || n->vert[2] == vert[1] || n->vert[2] == vert[2]) matches++;

	return (matches==2);
}


//********************************************************************************************
//*************************************************************************************** ADDITION
//********************************************************************************************

void MODEL::add_edge(VERT *p0, VERT *p1)
{
	EDGE *e = new EDGE;

	if (!last_edge)
	{
		first_edge = e;
	}
	else
	{
		last_edge->next = e;
	}
	last_edge = e;
	e->next = NULL;
	e->end[0] = p0;
	e->end[1] = p1;
	edge_count++;
}
VERT *MODEL::add_vert(double *p)
{
	int index;
	if (last_vert) index = last_vert->id + 1;
	else index = 0;
	return add_vert(index,p);
}
VERT *MODEL::add_vert(int index)
{
	double p[3] = {0,0,0};
	return add_vert(index, p);
}
VERT *MODEL::add_vert(int index, double *p)
{
	VERT *v = new VERT;
	v->pos[0] = p[0];
	v->pos[1] = p[1];
	v->pos[2] = p[2];

	if (!last_vert)
	{
		first_vert = v;
	}
	else
	{
		last_vert->next = v;
	}
	last_vert = v;
	v->id = index;
	v->src_model = this;
	v->next = NULL;


	vert_count++;
	return v;
}
//Remove a polygon from the adj list - happens if this polygon is deleted
//ADDITIONALLY! Remove from adj_p as well!
void VERT::remove_adj(POLY *poly)
{
	int i,j;
	for (i=0; i<num_adj; i++)
	{
		if (adj[i]==poly)
		{
			for (j=i;j<num_adj-1;j++)
			{
				adj[j] = adj[j+1];
			}
			adj[j] = NULL;
			num_adj--;
			break;
		}
	}

	for (i=0; i<num_adj_p; i++)
	{
		if (poly->vert[0]==adj_p[i] || poly->vert[1]==adj_p[i] || poly->vert[2]==adj_p[i])
		{
			for (j=i;j<num_adj_p-1;j++)
			{
				adj_p[j] = adj_p[j+1];
			}
			adj_p[j] = NULL;
			num_adj_p--;
			i--;
		}
	}
}
//Add a list of POLYGONS adjacent to a vertex
void VERT::add_adj(POLY *poly)
{
	if (num_adj+1>=MAX_ADJ)
	{
		frame.error("Too many adjacent polys on vertex: %i requested where %i are available.",num_adj+2,MAX_ADJ);
		return;
	}
	adj[num_adj] = poly;
	num_adj++;
}
//This is confusing: This is actually a list of adjacent VERTICES
void VERT::add_adj_p(VERT *p)
{
	if (num_adj_p+1>=MAX_ADJ)
	{
		//This should never happen - it will fail on add_adj() first, before failing on this one
		frame.error("Too many adjacent points on vertex: %i requested where %i are available.",num_adj_p+2,MAX_ADJ);
		return;
	}

	//Make sure we do not add duplicates!
	for (int i=0; i<num_adj_p; i++)
	{
		if (adj_p[i]==p)
			return;
	}

	adj_p[num_adj_p] = p;
	num_adj_p++;
}
void POLY::add_vert(int i, VERT *p)
{
	if (p->pos[0]<X0) X0 = p->pos[0];
	if (p->pos[0]>X1) X1 = p->pos[0];
	if (p->pos[1]<Y0) Y0 = p->pos[1];
	if (p->pos[1]>Y1) Y1 = p->pos[1];
	if (p->pos[2]<Z0) Z0 = p->pos[2];
	if (p->pos[2]>Z1) Z1 = p->pos[2];

	vert[i] = p;
}
POLY *MODEL::add_poly(int vert_count, VERT *one, ...)
{
	POLY *new_poly;

	new_poly = new POLY(vert_count);
	new_poly->id = ++last_poly_id;
	int i;
	VERT *e;
	va_list marker;
	va_start( marker, vert_count );
	for (i=0; i<vert_count; i++)
	{
		e = va_arg( marker, VERT*);
		new_poly->add_vert(i,e);
		new_poly->src_model = this;
	}
	va_end( marker );

	if (first_poly==NULL)
	{
		first_poly = new_poly;
		last_poly = new_poly;
	}
	else
	{
		last_poly->next = new_poly;
		new_poly->prev = last_poly;
		last_poly = new_poly;
	}
	poly_count++;
	new_poly->set_longest_edge();
	return new_poly;
}

void MODEL::remove_poly(POLY *poly)
{
	poly->vert[0]->remove_adj(poly);
	poly->vert[1]->remove_adj(poly);
	poly->vert[2]->remove_adj(poly);

	if (poly->bin_src)
	{
		if (poly==poly->bin_src->first_poly)	poly->bin_src->first_poly = poly->bin_next;
		if (poly==poly->bin_src->last_poly)		poly->bin_src->last_poly = poly->bin_prev;
	}
	if (poly->bin_next)							poly->bin_next->bin_prev = poly->bin_prev;
	if (poly->bin_prev)							poly->bin_prev->bin_next = poly->bin_next;
	poly->bin_next = poly->bin_prev = NULL;


	if (poly==first_poly)	first_poly = poly->next;
	if (poly==last_poly)	last_poly = poly->prev;
	if (poly->next)			poly->next->prev = poly->prev;
	if (poly->prev)			poly->prev->next = poly->next;
	poly->next = poly->prev = NULL;


	delete poly;

	poly_count--;
}



//********************************************************************************************
//*************************************************************************************** DRAWING
//********************************************************************************************

//Draw entire model
void MODEL::draw()
{
	draw(0);
}
void MODEL::draw(int vp)
{
	if (!ready) return;

//	glDrawElements(GL_TRIANGLES, poly_count*3, GL_UNSIGNED_INT, va_indices);

	if (display_list[vp]<=0)
	{
		//List is bad!  Recreate!
		create_list(vp);
	}

	//draw list
	glCallList(display_list[vp]);
}

void MODEL::create_all_lists()
{
	create_list(0);create_list(1);create_list(2);create_list(3);
}
void MODEL::create_list()
{
	//If called without VP, create only active
	create_list(frame.control->active_viewport);
}
//Store model in display list
void MODEL::create_list(int vp)
{
	if (!ready) return;

	if (display_list[vp]>0)
	{
		//destroy current list
		glDeleteLists(display_list[vp],1);
		display_list[vp] = -1;
	}

	display_list[vp] = glGenLists(1);
	glNewList(display_list[vp],GL_COMPILE);
	{
		glEnable(GL_CULL_FACE);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1,1);

		if (frame.display.viewports.vp[vp].get_hidden())
			glColorMask(false,false,false,false);

		if (frame.display.viewports.vp[vp].get_poly())
			draw_polys(vp);

		if (frame.display.viewports.vp[vp].get_hidden())
			glColorMask(true,true,true,true);

		if (frame.display.viewports.vp[vp].get_wire())		draw_edges();
		if (frame.display.viewports.vp[vp].get_normal())	draw_normals();
		if (frame.display.viewports.vp[vp].get_points())	draw_points();

		glDisable(GL_POLYGON_OFFSET_FILL);
	}
	glEndList();
}
//Draw all polys in a model
void MODEL::draw_polys(int vp)
{
	VERT *v = first_vert;
	double d;
	while (v)
	{
		//Temporary visualization options - cycle through them with key defined in control.cpp
		// Each renders the polygons of a model based on some given data about either the
		// polys themselves or the vertices of which they are composed
		switch (frame.display.viewports.vp[vp].get_visualization())
		{
		default:
		case 0:		//Brown with moss residue
			{
				v->color[0] = .40f;
				v->color[1] = .25f;
				v->color[2] = .10f;
				v->color[1] += v->residue.moss * .5f;
				break;
			}
		case 1:		//Normals
			{
				v->color[0] = v->normal[0];
				v->color[1] = v->normal[1];
				v->color[2] = v->normal[2];
				break;
			}
		case 2:		//Curvature
			{
				d = v->curvature * (1.0/0.00370919);
				double p[3] = {.05,.01,.001};
				v->color[0]=v->color[1]=v->color[2] = 0;
				if (d>p[0])			v->color[0] = .5 + ((d-p[0]) / (1-p[0])) / .5;
				else if (d>.01)		v->color[1] = .5 + ((d-p[1]) / (1-p[1])) / .5;
				else if (d>.001)	v->color[2] = .5 + ((d-p[2]) / (1-p[2])) / .5;
				break;
			}
		case 3:		//Offset magnitude
			{
				d = sqrt(v->residue.offset[0]*v->residue.offset[0] + v->residue.offset[1]*v->residue.offset[1] + 
					v->residue.offset[2]*v->residue.offset[2]);
				d *= 1000;
				v->color[0] = 1 - d;
				v->color[1] = 1 - d;
				v->color[2] = 1 - d;
				break;
			}
		case 4:		//Offset direction
			{
				d = 1000;
				v->color[0] = fabs(v->residue.offset[0])*d;
				v->color[1] = fabs(v->residue.offset[1])*d;
				v->color[2] = fabs(v->residue.offset[2])*d;
				break;
			}
		case 5:		//Collision count
			{
				d = v->residue.collisions / 100.0;
				v->color[0] = d;
				v->color[1] = 1 - d;
				v->color[2] = 0;
			}

		}



		v = v->next;
	}

	if (frame.display.viewports.vp[vp].get_lighting())
	{
		glColorMaterial(GL_FRONT,GL_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);
	}

	POLY *current = first_poly;
	glBegin(GL_TRIANGLES);
	while (current)
	{
		current->draw();
		current = current->next;
	}
	glEnd();

	glDisable(GL_COLOR_MATERIAL);
	glDisable(GL_LIGHTING);
}
//Draw all edges in a model
void MODEL::draw_edges()
{
	EDGE *current;

	glColor4f(1,1,1,.25f);

	glBegin(GL_LINES);
	current = first_edge;
	while (current)
	{
		glVertex3dv(current->end[0]->pos);
		glVertex3dv(current->end[1]->pos);
		current = current->next;
	}
	glEnd();
}
//Draw all normals in a model
void MODEL::draw_normals()
{
	double *p;
	double *n;
	float *c;
	VERT *P;
	glBegin(GL_LINES);
	P = first_vert;
	while (P)
	{
		if (!P->orphan)
		{
			p = P->pos;
			n = P->dnormal;
			c = P->color;
			glColor4f(c[2],c[0],c[1],1);	glVertex3dv(p);
			glColor4f(c[2],c[0],c[1],0);	glVertex3d(p[0]+n[0],p[1]+n[1],p[2]+n[2]);
		}
		P = P->next;
	}
	glEnd();
}
//Draw all points in a model
void MODEL::draw_points()
{
	VERT *P;
	glPointSize(3);

	glBegin(GL_POINTS);
	P = first_vert;
	while (P)
	{
		if (!P->orphan)
		{
			if (P->marked)	glColor4f(0,1,0,1);
			else			glColor4f(1,.5f,0,1);
			glVertex3dv(P->pos);
		}
		P = P->next;
	}
	glEnd();

	glPointSize(1);
}
//Draw a given polygon.  We assume they're all triangles,
// hence the loop is unrolled completely
void POLY::draw()
{
	VERT *v;

	//I have disabled rendering of marked polys differently for now

//	if (!subdiv_needed && !marked)
	{
		v = vert[0];
		glNormal3dv(v->normal);
		glColor4fv(v->color);
		glVertex3dv(v->pos);

		v = vert[1];
		glNormal3dv(v->normal);
		glColor4fv(v->color);
		glVertex3dv(v->pos);

		v = vert[2];
		glNormal3dv(v->normal);
		glColor4fv(v->color);
		glVertex3dv(v->pos);

	}
/*	else if (subdiv_needed)
	{
		glColor4f(0,1,0,1);

		v = vert[0];
		glNormal3dv(v->normal);
		glVertex3dv(v->pos);

		v = vert[1];
		glNormal3dv(v->normal);
		glVertex3dv(v->pos);

		v = vert[2];
		glNormal3dv(v->normal);
		glVertex3dv(v->pos);
	}
	else
	{
		glColor4f(1,0,0,1);

		v = vert[0];
		glNormal3dv(v->normal);
		glVertex3dv(v->pos);

		v = vert[1];
		glNormal3dv(v->normal);
		glVertex3dv(v->pos);

		v = vert[2];
		glNormal3dv(v->normal);
		glVertex3dv(v->pos);
	}//*/
}


//********************************************************************************************
//***************************************************************************************** ETC
//********************************************************************************************

//Return a vertex, given it's index (order it was added).
// Excruciatingly, painfully, monstrously slow.  Don't use it.
VERT *MODEL::get_vert(int index)
{
	VERT *v = first_vert;
	for (int i=0; i<index; i++)
	{
		v = v->next;
		if (!v)
		{
			frame.error("Could not get vert %i, not found.",i);
			break;
		}
	}
	return v;
}
//Shoves all vertices in some direction governed by their normal and
// the input scale.  Then it recalculates normals.
void MODEL::translate_by_normals(float scale)
{
	if (!ready) return;

	VERT *current = first_vert;
	scale *= 0.001f;

	while (current)
	{
		current->pos[0] += current->normal[0] * scale;
		current->pos[1] += current->normal[1] * scale;
		current->pos[2] += current->normal[2] * scale;
		current = current->next;
	}


	recalculate();
}


void MODEL::vector_field_test()
{

	VERT *v;
	double rx, ry, rz;

	float scale = .01f;

	v = first_vert;
	while (v)
	{
		frame.vector_field->get_vector(v->pos[0],v->pos[1],v->pos[2],&rx,&ry,&rz);

		rx *= scale; ry *= scale; rz *= scale;

		// When we have a change sweeping an entire model, we should do the add_bins
		//after we are done with all elements - no need to adjust bins until then.
		//This is faster than calling move() each time now.

		v->pos[0] += rx;v->pos[1] += ry;v->pos[2] += rz;
		v=v->next;
	}

	init_bins();
	recalculate();
}

//Subdivide a given triangle.  Not sure where this one is going yet.
void POLY::subdivide()
{
	int i, j, i0, i1, i2, x, x1;
	VERT *v[3];
	int which;

	if (!subdiv_needed)
	{
		//If this poly is adjacent to one that is subdivided, then we MUST
		// subdivide this one, but in a limited sense.  We must simply utilize
		// every shared new vertex.  As a result, we WILL divide into either
		// 2, 3, or 4 polygons, definitely.
		//The issue is deciding which ways to slice it.

		bool n[3] = {false,false,false};
		if (neighbor[0]) n[0] = neighbor[0]->subdiv_needed;
		if (neighbor[1]) n[1] = neighbor[1]->subdiv_needed;
		if (neighbor[2]) n[2] = neighbor[2]->subdiv_needed;
		int N = n[0] + n[1] + n[2];	//yay for adding bools

		if (n[0] || n[1] || n[2])
		{
			//Decide which verts need to be created.
			// This is different than for pure subdivided polys - it may not need 3 verts
			for (i=0; i<3; i++)
			{
				if (n[i])
				{
					if (neighbor[i]->subdiv_performed)
					{
						//If the neighbor is subdivided, it has already created a vertex on the current edge.
						//Figure out which neighbor will lend a vertex
						which = -2;
						if (neighbor[i]->neighbor[0]==this) which = 0;
						if (neighbor[i]->neighbor[1]==this) which = 1;
						if (neighbor[i]->neighbor[2]==this) which = 2;
						if (which>=0)	sub_vert[i] = neighbor[i]->sub_vert[which];
						else frame.error("Polygon neighbors not reciprocal, in subdivide()");
					}
					else
					{
						//Otherwise, make a new one and calculate its position
						sub_vert[i] = src_model->calc_new_vert(vert,i);
					}
					//Spread the residue
					sub_vert[i]->residue.interpolate(&vert[i]->residue,&vert[(i+1)%3]->residue,.5);
				}
			}

			//Decided how to slice it
			if (N==1)	//Only one neighboring polygon subdivided
			{
				//N now determines which edge will be divided
				if (n[0]) N=0; if (n[1]) N=1; if (n[2]) N=2;

				for (i=0; i<2; i++)
				{
					if (i==0) {i0 = (2 + N) % 3; i1 = N;}
					if (i==1) {i0 = (1 + N) % 3; i1 = (2 + N) % 3;}
					v[0] = vert[i0];
					v[1] = vert[i1];
					v[2] = sub_vert[N];
					sub_poly[i] = src_model->add_poly(3,v[0],v[1],v[2]);
					v[0]->add_adj(sub_poly[i]);
					v[1]->add_adj(sub_poly[i]);
					v[2]->add_adj(sub_poly[i]);
				}
			}
			else if (N==2)	//Two neighboring polygons subdivided -- worst case!
			{
				//N now determines which edge to NOT divide
				if (n[0]&&n[1]) N=0;
				if (n[1]&&n[2]) N=1;
				if (n[2]&&n[0]) N=2;

				//If you REALLY must understand this, I can show you the paper on which I worked this out.
				// It does not make a lot of sense here.
				x = N;
				x1 = (N+1) % 3;
				i0 = (N+1) % 3;
				i1 = (N+2) % 3;
				i2 = N;
				for (i=0; i<3; i++)
				{
					if (i==0)
					{
						v[0] = sub_vert[x1];
						v[1] = sub_vert[x];
						v[2] = vert[i0];
					}
					if (i==1)
					{
						v[0] = sub_vert[x];
						v[1] = sub_vert[x1];
						v[2] = vert[i1];
					}
					if (i==2)
					{
						v[0] = sub_vert[x];
						v[1] = vert[i1];
						v[2] = vert[i2];
					}

					sub_poly[i] = src_model->add_poly(3,v[0],v[1],v[2]);
					v[0]->add_adj(sub_poly[i]);
					v[1]->add_adj(sub_poly[i]);
					v[2]->add_adj(sub_poly[i]);
				}
			}
			else if (N==3)	//Three neighboring polygons subdivided -- easiest case!
			{
				//Just subdivide into four!  Simple!
				for (i=0; i<4; i++)
				{
					if (i==0) {v[0] = vert[0]; v[1] = sub_vert[0]; v[2] = sub_vert[2];}
					if (i==1) {v[0] = sub_vert[0]; v[1] = vert[1]; v[2] = sub_vert[1];}
					if (i==2) {v[0] = sub_vert[2]; v[1] = sub_vert[1]; v[2] = vert[2];}
					if (i==3) {v[0] = sub_vert[0]; v[1] = sub_vert[1]; v[2] = sub_vert[2];}
					sub_poly[i] = src_model->add_poly(3,v[0],v[1],v[2]);
					v[0]->add_adj(sub_poly[i]);
					v[1]->add_adj(sub_poly[i]);
					v[2]->add_adj(sub_poly[i]);
				}
			}
			else
				frame.error("The logic in MODEL::subdivide() is wonky; talk to Doug about it.");

			subdiv_performed = true;
		}
	}
	else
	{
		//This polygon will be subdivided into exactly four sub polys.
		// Three verts will need to be used to do this.  However, an adjacent
		// poly may have already created a shared vertex, if it has already
		// been subdivided.  If so, figure out which one.  If not, create a new
		// vertex.
		//The four new polys will be created with these vertices, once specified.

		//Decide which verts need to be created.
		for (i=0; i<3; i++)
		{
			which = -1;
			if (neighbor[i] && neighbor[i]->subdiv_performed)
			{
				which = -2;
				if (neighbor[i]->neighbor[0]==this) which = 0;
				if (neighbor[i]->neighbor[1]==this) which = 1;
				if (neighbor[i]->neighbor[2]==this) which = 2;
				if (which<0) frame.error("Polygon neighbors not reciprocal, in subdivide()");
			}
			if (which<0)
			{
				//Create your own if neighbor didn't subdivide on that edge
				sub_vert[i] = src_model->calc_new_vert(vert,i);
			}
			else
			{
				//Steal neighbor's vertex if you can
				sub_vert[i] = neighbor[i]->sub_vert[which];
			}
		}

		//Create new polys
		for (i=0; i<4; i++)
		{
			//Create each new poly

			if (i==0) {v[0] = vert[0]; v[1] = sub_vert[0]; v[2] = sub_vert[2];}
			if (i==1) {v[0] = sub_vert[0]; v[1] = vert[1]; v[2] = sub_vert[1];}
			if (i==2) {v[0] = sub_vert[2]; v[1] = sub_vert[1]; v[2] = vert[2];}
			if (i==3) {v[0] = sub_vert[0]; v[1] = sub_vert[1]; v[2] = sub_vert[2];}

			sub_poly[i] = src_model->add_poly(3,v[0],v[1],v[2]);
			v[0]->add_adj(sub_poly[i]);
			v[1]->add_adj(sub_poly[i]);
			v[2]->add_adj(sub_poly[i]);

			sub_poly[i]->subdiv_levels = subdiv_levels - 1;
		}



		//Move by normals, used for testing
		//Temporarily disabled
		if (false)
		{
			double s = 3000;
			v[0]->move(normal[0]/s,normal[1]/s,normal[2]/s);
			v[1]->move(normal[0]/s,normal[1]/s,normal[2]/s);
			v[2]->move(normal[0]/s,normal[1]/s,normal[2]/s);
		}

		//Spread the curvature/residue and such
		for (i=0; i<3; i++)
		{
			if (sub_vert[i])
			{
				j = (i+1)%3;
				sub_vert[i]->normal[0] = (vert[i]->normal[0] + vert[j]->normal[0]) / 2.0;
				sub_vert[i]->normal[1] = (vert[i]->normal[1] + vert[j]->normal[1]) / 2.0;
				sub_vert[i]->normal[2] = (vert[i]->normal[2] + vert[j]->normal[2]) / 2.0;
				sub_vert[i]->curvature = (vert[i]->curvature + vert[j]->curvature) / 2.0;
				sub_vert[i]->residue.interpolate(&vert[i]->residue,&vert[j]->residue,.5);
			}
		}

		subdiv_performed = true;
	}
}


VERT *MODEL::calc_new_vert(VERT **vert, int i)
{
	double p[3];
	int i0 = i;
	int i1 = (i+1) % 3;

	//Linear subdivision!  Creates ugly facets
	p[0] = ((vert[i0]->pos[0] - vert[i1]->pos[0]) / 2.0) + vert[i1]->pos[0];
	p[1] = ((vert[i0]->pos[1] - vert[i1]->pos[1]) / 2.0) + vert[i1]->pos[1];
	p[2] = ((vert[i0]->pos[2] - vert[i1]->pos[2]) / 2.0) + vert[i1]->pos[2];




	//If we'd like, we can modify p[] to reflect curvature,
	// rather than linearly interpolating from vert[i0-i1].
	//HOWEVER -- my curvature calculations seem to be fairly wrong.
	// As a result, this causes the models to basically explode.
	// So, until curvature is addressed better, this will have
	// to be ignored.
/*
	float c = vert[i0]->curvature + vert[i1]->curvature;
	c /= 2;
	float d = 1;
	double n[3];
	n[0] = (vert[i0]->normal[0] + vert[i1]->normal[0]) / 2.0;
	n[1] = (vert[i0]->normal[1] + vert[i1]->normal[1]) / 2.0;
	n[2] = (vert[i0]->normal[2] + vert[i1]->normal[2]) / 2.0;
	p[0] += c * d * n[0];
	p[1] += c * d * n[1];
	p[2] += c * d * n[2];
*/


	return add_vert(p);
}

//For each polygon in the model, subdivide polygons marked as subdiv_needed
void MODEL::subdivide()
{
	if (!ready)
		return;

	err_log << "Subdividing..." << endl;

	POLY *poly;

	int n = 1;
	frame.display.overlay->set_wait_reason("Subdividing");
	int static_count = poly_count;

	poly = first_poly;
	while (poly)
	{
		if (n%100==0)
			frame.display.overlay->post_progress((float)n/(float)static_count);

		//Create new polys for each existing triangle, if needed.
		// Store new polys in sub_poly[4].
		// Store new verts in sub_vert[3].
		// Verts are shared between polys that are subdividing.
		//Once we have looped, just remove all the original polygons
		// that were just subdivided - they have been overlapped with
		// the new smaller polys.  We keep the old polys around during
		// the first loop so that we can recognize shared vertices.
		//QUESTION: Is sub_poly[4] even used anywhere?  Sort of.
		// It is used to pass to add_adj.  Of course, just a single
		// *poly would do it.  But, it is nice to keep track of it.

		poly->subdivide();

		poly = poly->next;
		n++;
	}
	err_log << "Subdivision main loop complete.  Performing cleanup..." << endl;

	n = 0;
	bool loop = false;
	POLY *remove;
	poly = first_poly;
	while (poly)
	{
		//Determine if further subdivision is needed
		if (!poly->subdiv_performed && poly->subdiv_levels>0)
		{
			loop = true;
			poly->subdiv_needed = true;
		}

		remove = poly;
		poly = poly->next;
		n++;

		//Remove a polygon that has been subdivided
		if (remove->subdiv_performed)
		{
			remove_poly(remove);
		}
	}



err_log << " POST SUBDIVISION Parse edges:" << endl;
	parse_edges();
err_log << " POST SUBDIVISION Parse polys:" << endl;
	parse_polys();



	//If a polygon must be subdivided more, then its neighbors try to get subdivided as well.
	// This makes for a more even transition.
	poly = first_poly;
	while (poly)
	{
		if (poly->subdiv_levels>0)
		{
			if (poly->neighbor[0] && !poly->neighbor[0]->subdiv_needed)
			{
				poly->neighbor[0]->subdiv_needed = true;
				poly->neighbor[0]->subdiv_levels = 1;
			}
			if (poly->neighbor[1] && !poly->neighbor[1]->subdiv_needed)
			{
				poly->neighbor[1]->subdiv_needed = true;
				poly->neighbor[1]->subdiv_levels = 1;
			}
			if (poly->neighbor[2] && !poly->neighbor[2]->subdiv_needed)
			{
				poly->neighbor[2]->subdiv_needed = true;
				poly->neighbor[2]->subdiv_levels = 1;
			}
		}
		poly = poly->next;
	}

	//If any polys still need subdivision and we're not doing it incrementally, do it again!
	if (loop && !frame.config.get_bool(VAR_INCREMENTAL_SUBDIV))
		subdivide();
}

void POLY::set_subdiv_needed(double r, double rad_limit)
{
	//This is how NOT to decide the levels:
//	subdiv_levels = (int)(r / rad_limit);

	//Correct method: polygon edges will be cut in half
	// for each subdivision
	int i = 0;
	while (r > rad_limit)
	{
		r /= 2;
		i++;
	}

	//If we already decided it needs subdivision, preserve it!
	// We don't want it to go any lower, only higher.
	if (i>subdiv_levels)			subdiv_levels = i;
	if (subdiv_levels>0)			subdiv_needed = true;

	sub_poly[0]=sub_poly[1]=sub_poly[2]=sub_poly[3] = NULL;
	sub_vert[0]=sub_vert[1]=sub_vert[2] = NULL;
}

//Recalculate stuff.  Use after modifying positions and such.
void MODEL::recalculate()
{
	recalculate(true);
}
void MODEL::recalculate(bool new_bins)
{
	if (!ready) return;

	err_log << "*** Recalculating model..." << endl;

frame.display.overlay->show_calculating("Calculate poly normals");
err_log << " Init poly normals:" << endl;
	init_poly_normals();
frame.display.overlay->show_calculating("Calculate vert normals");
err_log << " Init vert normals:" << endl;
	init_vert_normals();
frame.display.overlay->show_calculating("Calculate curvature");
err_log << " Calculate curvature:" << endl;
	calculate_curvature_laplace();
//	calculate_curvature_desbrun();


	if (new_bins)
	{
frame.display.overlay->show_calculating("Init bins");
		err_log << " Init bins:" << endl;
			init_bins();
	}


frame.display.overlay->show_calculating("Create all lists");
err_log << " Create all lists: " << endl;
	create_all_lists();

//	frame.vertex_array.create();


	//This code block can be used to test the number of adjacent polys created
/*
	int a = -1;
	int b = -1;
	VERT *v = first_vert;
	while (v)
	{
		if (v->num_adj>a) a = v->num_adj;
		if (v->num_adj_p>b) b = v->num_adj_p;
		v = v->next;
	}
	err_log << "MAX ADJ:   " << a << endl;
	err_log << "MAX ADJ_P: " << b << endl;
*/

	err_log << "Recalculation complete." << endl;
}

void VERT::move(double *m)
{
	move(m[0],m[1],m[2]);
}
//Move a vertex and add its polys back into the bins
void VERT::move(double mx, double my, double mz)
{
	residue.offset[0] += mx;
	residue.offset[1] += my;
	residue.offset[2] += mz;

	POLY *poly;
	int i;
	int p[3];
	pos[0] += mx;
	pos[1] += my;
	pos[2] += mz;

	//get bin pos. & See if it's in new bin (super quick)
	for (i=0;i<num_adj;i++)
	{
		poly = adj[i];
		if (frame.bins->new_bin_needed(poly,p))
		{
			//Could be one, none, or each!  Don't know yet!
			if (poly->vert[0] == this)	//If this vert is the zeroth vert of the given poly
			{
				//Target poly p's vert[0] is self, thus this poly must be re-binlocated
				frame.bins->force_add_poly(p,poly);
			}
		}
	}
}


//********************************************************************************************
//******************************************************************************** Vertex Array Crap
//********************************************************************************************


//This vertex array stuff isn't so good - it renders slightly faster but
// requires much more pre-computation.
/*
VERTEX_ARRAY::VERTEX_ARRAY()
{
	ready = false;
	clear();
}
void VERTEX_ARRAY::clear()
{
	if (display_list>0) glDeleteLists(display_list,1);
	display_list = -1;
	if (ready)
	{
		free(verts);
		free(color);
		free(indices);
	}
	vc = 0;
	pc = 0;
	ready = false;
}
void VERTEX_ARRAY::create()
{

	if (ready) clear();

	int i;
	vc=0;
	pc=0;

	for (i=0; i<MAX_MODELS; i++)
	{
		if (frame.model[i].is_ready())
		{
			vc += frame.model[i].get_vert_count();
			pc += frame.model[i].get_poly_count();
		}
	}


	verts = (GLdouble*)calloc(vc*3,sizeof(GLdouble));
	color = (GLfloat*)calloc(vc*3,sizeof(GLfloat));
	normal = (GLdouble*)calloc(vc*3,sizeof(GLdouble));
	indices = (GLuint*)calloc(pc*3,sizeof(GLuint));

 
	int vi = 0;
	int pi = 0;
	for (i=0; i<MAX_MODELS; i++)
	{
		frame.model[i].contribute_va(verts,color,normal,indices,&vi,&pi);
	}

	glEnableClientState(GL_VERTEX_ARRAY);	glVertexPointer(3, GL_DOUBLE, 0, verts);
	glEnableClientState(GL_COLOR_ARRAY);	glColorPointer(3, GL_FLOAT, 0, color);
//	glEnableClientState(GL_NORMAL_ARRAY);	glNormalPointer(GL_double, 0, normal);



	if (display_list>0)
	{
		glDeleteLists(display_list,1);
		display_list = -1;
	}
	display_list = glGenLists(1);
	glNewList(display_list,GL_COMPILE);
	{
		if (frame.config.get_bool(VAR_HIDDEN_LINE))
			glColorMask(false,false,false,false);

		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1,1);

		if (frame.config.get_bool(VAR_DPOLY))
			glDrawElements(GL_TRIANGLES, pc*3, GL_UNSIGNED_INT, indices);

		if (frame.config.get_bool(VAR_HIDDEN_LINE))
			glColorMask(true,true,true,true);

		glDisable(GL_POLYGON_OFFSET_FILL);
	}
	glEndList();


	ready = true;

}
*/
/*
void MODEL::contribute_va(GLdouble *va_verts, GLfloat *va_color, GLdouble *va_normal, GLuint *va_indices, int *vi, int *pi)
{
	if (!ready) return;

	VERT *v = first_vert;
	POLY *p = first_poly;

	int vo = *vi / 3;
	int po = *pi;

	//Store information per vertex
	while (v)
	{
		va_verts[*vi] = v->pos[0];		va_verts[*vi+1] = v->pos[1];		va_verts[*vi+2] = v->pos[2];
		va_color[*vi] = v->color[0];	va_color[*vi+1] = v->color[1];		va_color[*vi+2] = v->color[2];
//		va_normal[*vi] = v->normal[0];	va_normal[*vi+1] = v->normal[1];	va_normal[*vi+2] = v->normal[2];

		*vi += 3;
		v = v->next;
	}

	//Store information per polygon
	while (p)
	{
		va_indices[*pi] = p->vert[0]->id + vo;
		va_indices[*pi+1] = p->vert[1]->id + vo;
		va_indices[*pi+2] = p->vert[2]->id + vo;

		*pi += 3;
		p = p->next;
	}
}
*/
/*
void VERTEX_ARRAY::draw()
{

	glCallList(display_list);
}
*/

//********************************************************************************************
//************************************************************************************** 
//********************************************************************************************


void MODEL::full_subdivision()
{
	err_log << "Performing full subdivision on model..." << endl;

	POLY *poly = first_poly;

	err_log << "POLY: " << first_poly << endl;
	while (poly)
	{
		err_log << "POLY: " << poly << endl;
		poly->subdiv_needed = true;
		poly->subdiv_levels = 1;
		poly = poly->next;
	}
	subdivide();
}


void MODEL::calculate_curvature_laplace()	//translation of taubin_laplace()
{
	int v, i;
	double total_curve = 0.0;
	double vert_laplace2[3];
	double smallest_c = 100.0, largest_c = 0.0, cur_curve;
	VERT *vert;
	POLY *poly;

	vert = first_vert;
	v = 0;
	while (vert)
	{
		// Initialize result of first Laplacian.
		vert->laplace[0]=vert->laplace[1]=vert->laplace[2] = 0;
			
		// Apply first Laplacian operator.
		for (i=0; i<vert->num_adj; i++)
		{
			poly = vert->adj[i];
			if (poly->vert[0] == vert)
			{
				vert->laplace[0] += poly->vert[2]->pos[0] - vert->pos[0];
				vert->laplace[1] += poly->vert[2]->pos[1] - vert->pos[1];
				vert->laplace[2] += poly->vert[2]->pos[2] - vert->pos[2];
			}
			else if (poly->vert[1] == vert)
			{
				vert->laplace[0] += poly->vert[0]->pos[0] - vert->pos[0];
				vert->laplace[1] += poly->vert[0]->pos[1] - vert->pos[1];
				vert->laplace[2] += poly->vert[0]->pos[2] - vert->pos[2];
			}
			else if (poly->vert[2] == vert)
			{
				vert->laplace[0] += poly->vert[1]->pos[0] - vert->pos[0];
				vert->laplace[1] += poly->vert[1]->pos[1] - vert->pos[1];
				vert->laplace[2] += poly->vert[1]->pos[2] - vert->pos[2];
			}
		}

		vert->laplace[0] /= vert->num_adj;
		vert->laplace[1] /= vert->num_adj;
		vert->laplace[2] /= vert->num_adj;

		vert = vert->next;
		v++;
	}

	// Apply second Laplacian operator.
	vert = first_vert;
	v = 0;
	while (vert)
	{
		// Initialize result of second Laplacian operator
		vert_laplace2[0] = 0.0;
		vert_laplace2[1] = 0.0;
		vert_laplace2[2] = 0.0;

		for (i=0; i<vert->num_adj; i++)
		{
			poly = vert->adj[i];

			if (poly->vert[0] == vert)
			{
				vert_laplace2[0] += poly->vert[2]->laplace[0] - vert->laplace[0];
				vert_laplace2[1] += poly->vert[2]->laplace[1] - vert->laplace[1];
				vert_laplace2[2] += poly->vert[2]->laplace[2] - vert->laplace[2];
			}
			else if (poly->vert[1] == vert)
			{
				vert_laplace2[0] += poly->vert[0]->laplace[0] - vert->laplace[0];
				vert_laplace2[1] += poly->vert[0]->laplace[1] - vert->laplace[1];
				vert_laplace2[2] += poly->vert[0]->laplace[2] - vert->laplace[2];
			}
			else if (poly->vert[2] == vert)
			{
				vert_laplace2[0] += poly->vert[1]->laplace[0] - vert->laplace[0];
				vert_laplace2[1] += poly->vert[1]->laplace[1] - vert->laplace[1];
				vert_laplace2[2] += poly->vert[1]->laplace[2] - vert->laplace[2];
			}

			//The method below was presented in the original function, but seemed..  wrong.
			// I replaced it with the one above.  Both result in crap curvature though.
/*
			if (poly->vert[0] == vert)
			{
				vert_laplace2[0] += poly->vert[2]->laplace[0] - vert->laplace[0];
				vert_laplace2[1] += poly->vert[2]->laplace[0] - vert->laplace[0];
				vert_laplace2[2] += poly->vert[2]->laplace[0] - vert->laplace[0];
			}
			else if (poly->vert[1] == vert)
			{
				vert_laplace2[0] += poly->vert[0]->laplace[0] - vert->laplace[0];
				vert_laplace2[1] += poly->vert[0]->laplace[0] - vert->laplace[0];
				vert_laplace2[2] += poly->vert[0]->laplace[0] - vert->laplace[0];
			}
			else if (poly->vert[2] == vert)
			{
				vert_laplace2[0] += poly->vert[1]->laplace[0] - vert->laplace[0];
				vert_laplace2[1] += poly->vert[1]->laplace[0] - vert->laplace[0];
				vert_laplace2[2] += poly->vert[1]->laplace[0] - vert->laplace[0];
			}
*/

		}	
		vert_laplace2[0] /= vert->num_adj;
		vert_laplace2[1] /= vert->num_adj;
		vert_laplace2[2] /= vert->num_adj;

		// Accumulate to total curvature.
		cur_curve = sqrt(dot(vert_laplace2, vert_laplace2));
		total_curve += cur_curve;


		vert->curvature = cur_curve;

		if (false)
		{
			vert->color[0] = cur_curve * 1000.0;
			vert->color[0] = 0;
			vert->color[1] = 0;
			vert->color[2] = 0;

			double d = cur_curve * (1.0/0.00370919);

			double p[3] = {.05,.01,.001};

			if (d>p[0])		vert->color[0] = .5 + ((d-p[0]) / (1-p[0])) / .5;
			else if (d>.01)	vert->color[1] = .5 + ((d-p[1]) / (1-p[1])) / .5;
			else if (d>.001)	vert->color[2] = .5 + ((d-p[2]) / (1-p[2])) / .5;
		}

		if (cur_curve < smallest_c) smallest_c = cur_curve;
		if (cur_curve > largest_c) largest_c = cur_curve;

		v++;
		vert = vert->next;
	}
}

//Not fully translated
/*
void MODEL::calculate_curvature_desbrun()
{
	int v, i, j, f, f2, t2;
	// Allocate data structure to store the triangle indices associated with each vertex.
	int* parent_tri = (int*)calloc(50, sizeof(int));
	int num_tri;
	double total_curve = 0.0, curve_factor;
	double v1[3], v2[3], v3[3], v4[3], e[3], v1_cross[3], v4_cross[3];
	double vert_area, cur_curve;
	int total_num_tri = 0;

	POLY *poly;
	POLY *poly2;
	VERT *vert;

	vert = first_vert;
	for (v=0; v<nverts; v++)
	{
		//faces for current vertex are in vert->adj[]


		vert_area = 0.0;
		e[0] = 0.0; e[1] = 0.0; e[2] = 0.0;
		for (i=0; i<vert->num_adj; i++)
		{
//			f = parent_tri[i];
			poly = vert->adj[i];
			if (poly->vert[0] == v)
			{
				v1[0] = poly->vert[0]->pos[0] - poly->vert[2]->pos[0];
				v1[1] = poly->vert[0]->pos[1] - poly->vert[2]->pos[1];
				v1[2] = poly->vert[0]->pos[2] - poly->vert[2]->pos[2];

				v2[0] = poly->vert[1]->pos[0] - poly->vert[2]->pos[0];
				v2[1] = poly->vert[1]->pos[0] - poly->vert[2]->pos[0];
				v2[2] = poly->vert[1]->pos[0] - poly->vert[2]->pos[0];

				// All triangles in the list have v in common.  Now, find triangle
				//	that has 2nd vertex in common.
				t2 = -1;
				for (j=0; j<vert->num_adj; j++)
				{
//					f2 = parent_tri[j];
					poly2 = vert->adj[j];


					if (f2 == f) continue;


					if ((tlist[f2]->verts[0] == tlist[f]->verts[1]) || 
						(tlist[f2]->verts[1] == tlist[f]->verts[1]) || 
						(tlist[f2]->verts[2] == tlist[f]->verts[1]))
					{
						t2 = f2;
						break;
					}
				}

				if (t2 == -1) break;
				// Now, assign v3 and v4.
				if (tlist[t2]->verts[0] == v)
				{
					v3[0] = vlist[tlist[t2]->verts[2]]->x - vlist[tlist[t2]->verts[1]]->x;
					v3[1] = vlist[tlist[t2]->verts[2]]->y - vlist[tlist[t2]->verts[1]]->y;
					v3[2] = vlist[tlist[t2]->verts[2]]->z - vlist[tlist[t2]->verts[1]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[0]]->x - vlist[tlist[t2]->verts[1]]->x;
					v4[1] = vlist[tlist[t2]->verts[0]]->y - vlist[tlist[t2]->verts[1]]->y;
					v4[2] = vlist[tlist[t2]->verts[0]]->z - vlist[tlist[t2]->verts[1]]->z;
				}
				else if (tlist[t2]->verts[1] == v)
				{
					v3[0] = vlist[tlist[t2]->verts[0]]->x - vlist[tlist[t2]->verts[2]]->x;
					v3[1] = vlist[tlist[t2]->verts[0]]->y - vlist[tlist[t2]->verts[2]]->y;
					v3[2] = vlist[tlist[t2]->verts[0]]->z - vlist[tlist[t2]->verts[2]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[1]]->x - vlist[tlist[t2]->verts[2]]->x;
					v4[1] = vlist[tlist[t2]->verts[1]]->y - vlist[tlist[t2]->verts[2]]->y;
					v4[2] = vlist[tlist[t2]->verts[1]]->z - vlist[tlist[t2]->verts[2]]->z;
				}
				else
				{
					v3[0] = vlist[tlist[t2]->verts[1]]->x - vlist[tlist[t2]->verts[0]]->x;
					v3[1] = vlist[tlist[t2]->verts[1]]->y - vlist[tlist[t2]->verts[0]]->y;
					v3[2] = vlist[tlist[t2]->verts[1]]->z - vlist[tlist[t2]->verts[0]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[2]]->x - vlist[tlist[t2]->verts[0]]->x;
					v4[1] = vlist[tlist[t2]->verts[2]]->y - vlist[tlist[t2]->verts[0]]->y;
					v4[2] = vlist[tlist[t2]->verts[2]]->z - vlist[tlist[t2]->verts[0]]->z;
				}
			}		    
			else if (poly->vert[1] == v)
			{
				v1[0] = vlist[tlist[f]->verts[1]]->x - vlist[tlist[f]->verts[0]]->x;
				v1[1] = vlist[tlist[f]->verts[1]]->y - vlist[tlist[f]->verts[0]]->y;
				v1[2] = vlist[tlist[f]->verts[1]]->z - vlist[tlist[f]->verts[0]]->z;

				v2[0] = vlist[tlist[f]->verts[2]]->x - vlist[tlist[f]->verts[0]]->x;
				v2[1] = vlist[tlist[f]->verts[2]]->y - vlist[tlist[f]->verts[0]]->y;
				v2[2] = vlist[tlist[f]->verts[2]]->z - vlist[tlist[f]->verts[0]]->z;

				// All triangles in the list have v in common.  Now, find triangle
				//	that has 2nd vertex in common.
				t2 = -1;
				for (j=0; j<num_tri; j++)
				{
					f2 = parent_tri[j];
					if (f2 == f) continue;
					if ((tlist[f2]->verts[0] == tlist[f]->verts[2]) || 
						(tlist[f2]->verts[1] == tlist[f]->verts[2]) || 
						(tlist[f2]->verts[2] == tlist[f]->verts[2]))
					{
						t2 = f2;
						break;
					}
				}

				if (t2 == -1) break;
				// Now, assign v3 and v4.
				if (tlist[t2]->verts[0] == v)
				{
					v3[0] = vlist[tlist[t2]->verts[2]]->x - vlist[tlist[t2]->verts[1]]->x;
					v3[1] = vlist[tlist[t2]->verts[2]]->y - vlist[tlist[t2]->verts[1]]->y;
					v3[2] = vlist[tlist[t2]->verts[2]]->z - vlist[tlist[t2]->verts[1]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[0]]->x - vlist[tlist[t2]->verts[1]]->x;
					v4[1] = vlist[tlist[t2]->verts[0]]->y - vlist[tlist[t2]->verts[1]]->y;
					v4[2] = vlist[tlist[t2]->verts[0]]->z - vlist[tlist[t2]->verts[1]]->z;
				}
				else if (tlist[t2]->verts[1] == v)
				{
					v3[0] = vlist[tlist[t2]->verts[0]]->x - vlist[tlist[t2]->verts[2]]->x;
					v3[1] = vlist[tlist[t2]->verts[0]]->y - vlist[tlist[t2]->verts[2]]->y;
					v3[2] = vlist[tlist[t2]->verts[0]]->z - vlist[tlist[t2]->verts[2]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[1]]->x - vlist[tlist[t2]->verts[2]]->x;
					v4[1] = vlist[tlist[t2]->verts[1]]->y - vlist[tlist[t2]->verts[2]]->y;
					v4[2] = vlist[tlist[t2]->verts[1]]->z - vlist[tlist[t2]->verts[2]]->z;
				}
				else
				{
					v3[0] = vlist[tlist[t2]->verts[1]]->x - vlist[tlist[t2]->verts[0]]->x;
					v3[1] = vlist[tlist[t2]->verts[1]]->y - vlist[tlist[t2]->verts[0]]->y;
					v3[2] = vlist[tlist[t2]->verts[1]]->z - vlist[tlist[t2]->verts[0]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[2]]->x - vlist[tlist[t2]->verts[0]]->x;
					v4[1] = vlist[tlist[t2]->verts[2]]->y - vlist[tlist[t2]->verts[0]]->y;
					v4[2] = vlist[tlist[t2]->verts[2]]->z - vlist[tlist[t2]->verts[0]]->z;
				}
			}
			else
			{
				v1[0] = vlist[tlist[f]->verts[2]]->x - vlist[tlist[f]->verts[1]]->x;
				v1[1] = vlist[tlist[f]->verts[2]]->y - vlist[tlist[f]->verts[1]]->y;
				v1[2] = vlist[tlist[f]->verts[2]]->z - vlist[tlist[f]->verts[1]]->z;

				v2[0] = vlist[tlist[f]->verts[0]]->x - vlist[tlist[f]->verts[1]]->x;
				v2[1] = vlist[tlist[f]->verts[0]]->y - vlist[tlist[f]->verts[1]]->y;
				v2[2] = vlist[tlist[f]->verts[0]]->z - vlist[tlist[f]->verts[1]]->z;

				// All triangles in the list have v in common.  Now, find triangle
				//	that has 2nd vertex in common.
				t2 = -1;
				for (j=0; j<num_tri; j++)
				{
					f2 = parent_tri[j];
					if (f2 == f) continue;
					if ((tlist[f2]->verts[0] == tlist[f]->verts[0]) || 
						(tlist[f2]->verts[1] == tlist[f]->verts[0]) || 
						(tlist[f2]->verts[2] == tlist[f]->verts[0]))
					{
						t2 = f2;
						break;
					}
				}

				if (t2 == -1) break;
				// Now, assign v3 and v4.
				if (tlist[t2]->verts[0] == v)
				{
					v3[0] = vlist[tlist[t2]->verts[2]]->x - vlist[tlist[t2]->verts[1]]->x;
					v3[1] = vlist[tlist[t2]->verts[2]]->y - vlist[tlist[t2]->verts[1]]->y;
					v3[2] = vlist[tlist[t2]->verts[2]]->z - vlist[tlist[t2]->verts[1]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[0]]->x - vlist[tlist[t2]->verts[1]]->x;
					v4[1] = vlist[tlist[t2]->verts[0]]->y - vlist[tlist[t2]->verts[1]]->y;
					v4[2] = vlist[tlist[t2]->verts[0]]->z - vlist[tlist[t2]->verts[1]]->z;
				}
				else if (tlist[t2]->verts[1] == v)
				{
					v3[0] = vlist[tlist[t2]->verts[0]]->x - vlist[tlist[t2]->verts[2]]->x;
					v3[1] = vlist[tlist[t2]->verts[0]]->y - vlist[tlist[t2]->verts[2]]->y;
					v3[2] = vlist[tlist[t2]->verts[0]]->z - vlist[tlist[t2]->verts[2]]->z;

					v4[0] = vlist[tlist[t2]->verts[1]]->x - vlist[tlist[t2]->verts[2]]->x;
					v4[1] = vlist[tlist[t2]->verts[1]]->y - vlist[tlist[t2]->verts[2]]->y;
					v4[2] = vlist[tlist[t2]->verts[1]]->z - vlist[tlist[t2]->verts[2]]->z;
				}
				else
				{
					v3[0] = vlist[tlist[t2]->verts[1]]->x - vlist[tlist[t2]->verts[0]]->x;
					v3[1] = vlist[tlist[t2]->verts[1]]->y - vlist[tlist[t2]->verts[0]]->y;
					v3[2] = vlist[tlist[t2]->verts[1]]->z - vlist[tlist[t2]->verts[0]]->z;
					
					v4[0] = vlist[tlist[t2]->verts[2]]->x - vlist[tlist[t2]->verts[0]]->x;
					v4[1] = vlist[tlist[t2]->verts[2]]->y - vlist[tlist[t2]->verts[0]]->y;
					v4[2] = vlist[tlist[t2]->verts[2]]->z - vlist[tlist[t2]->verts[0]]->z;
				}
			}
		    
			// NO NEED to normalize vectors.

			// Calculate contribution of triangle to the area.  Note
			//	that area must be calculated before normalization since
			//	area is calculated using cross product.
			vert_area += area (v1,v2);
		    
			// Apply cot in form of dot and cross-prods to edge vector. edge = v1-v2.
			cross(v1,v2,v1_cross);
			cross(v2,v1_cross,v1_cross);
			cross(v4,v3,v4_cross);
			cross(v3,v4_cross,v4_cross);

			curve_factor = (sqrt(dot(v1_cross,v1_cross))/sqrt(dot(v2,v2)))*(dot(v1,v2)/dot(v1,v1_cross)) + 
			(sqrt(dot(v4_cross,v4_cross))/sqrt(dot(v3,v3)))*(dot(v4,v3)/dot(v4,v4_cross));
			e[0] += curve_factor * (v1[0] - v2[0]);
			e[1] += curve_factor * (v1[1] - v2[1]);
			e[2] += curve_factor * (v1[2] - v2[2]); 
		    
			total_num_tri++;
		}

		if (t2 == -1) continue;;

		// Accumulate curvature from vertex v.
		cur_curve = sqrt(dot(e,e))/(4.0 *vert_area);
		cur_curve /= 10.0;
		total_curve += cur_curve;

		if (cur_curve < 0.0) printf ("Negative curvature.\n");

		// Assign color according to curvature value.
		vlist[v]->r = cur_curve;
		vlist[v]->g = 0.0;
		vlist[v]->b = 0.0;
		vlist[v]->a = 0.0; 

		//printf ("curvature %lf\n", cur_curve);

		vert = vert->next;
	}

	printf ("total num of triangles: %i\n", total_num_tri);
	return total_curve;
}
*/







