#include "main.h"
#include "subdivide.h"


extern LOGGER err_log;



//Arguments:
//num_times - number of times to run the subdivision scheme
//mask_size - the size of the desired mask; provide N when an NxNxN mask is desired
//            must be one of the following: 5, 7, 9, 11, 13, 15
//*coarse - provide a pointer to the beginning of the coarse vector field
//*fine - provide a pointer to the beginning of the (initally empty) fine vector field

int subdivide(int num_times, int mask_size, float *coarse, float *fine)
{
	printf("Subdivide: %i, %i\n",num_times, mask_size);


	err_log << "Geting mask." << endl;
    //Creates the mask based on the argument mask_size
    Mask3D m11;
    Mask3D m12;
	int i;

    switch (mask_size)
    {
        case 5:
            m11.recreate("data/d5x5x5x.msk");
            m12.recreate("data/od5x5x5.msk");
            break;
        case 7:
            m11.recreate("data/d7x7x7.msk");
            m12.recreate("data/od7x7x7.msk");
            break;
        case 9:
            m11.recreate("data/d9x9x9.msk");
            m12.recreate("data/od9x9x9.msk");
            break;
        case 11:
            m11.recreate("data/d11x11x11.msk");
            m12.recreate("data/od11x11x11.msk");
            break;
        case 13:
            m11.recreate("data/d13x13x13.msk");
            m12.recreate("data/od13x13x13.msk");
            break;
        case 15:
            m11.recreate("data/d15x15x15.msk");
            m12.recreate("data/od15x15x15.msk");
            break;
        default:
            printf("Error: mask size provided was %i. Mask size must be: 5, 7, 9, 11, 13 or 15\n", mask_size);
            return -1;
    }
    

    int vSizeX = 0;
    int vSizeY = 0; 
    int vSizeZ = 0;

    //Some explanation of *field: field is an array of VectorField3D's which holds each successive
    //iteration of the subdivision. Thus, field[0] holds the original coarse vector field provided
    //as input, field[1] holds the VectorField as it is after the first round of subdivision, and so
    //on up to field[num_times], which holds the field that was refined num_times, as specified by 
    //the argument to subdivide()
    VectorField3D *field = new VectorField3D[num_times+1];

    //Initializes the subdivision scheme
    FlowSubdiv3D scheme;
    scheme.setDiagonalMask(&m11);
    scheme.setOffDiagonalMask(&m12);

    vSizeX = (int) *coarse; coarse++;
    vSizeY = (int) *coarse; coarse++;
    vSizeZ = (int) *coarse; coarse++;

	//Checks to see that numbers are odd?
//	assert( vSizeX == 2*(vSizeX/2)+1 && vSizeY == 2*(vSizeY/2)+1 && vSizeZ == 2*(vSizeZ/2)+1);
	assert( vSizeX % 2 == 1 && vSizeY % 2 == 1 && vSizeZ % 2 == 1);



	err_log << "Filling coarse field." << endl;
    //Puts the coarse vector field in to field[0]
	int t = 0;
    field[0].reset(vSizeX, vSizeY, vSizeZ);
    for (i = -field[0].nx(); i <= field[0].nx(); i++)
        for (int j = -field[0].ny(); j <= field[0].ny(); j++)
            for (int k = -field[0].nz(); k <= field[0].nz(); k++)
            {
                field[0].valXRef(i,j,k) = *coarse; coarse++;
                field[0].valYRef(i,j,k) = *coarse; coarse++;
                field[0].valZRef(i,j,k) = *coarse; coarse++;
			}

    //Runs the subdivision scheme num_times times
    for (i = 0; i < num_times; i++)
    {
        vSizeX = (2 * vSizeX) - 1;
        vSizeY = (2 * vSizeY) - 1;
        vSizeZ = (2 * vSizeZ) - 1;
        VectorField3D big;
		err_log << "Calling subdivision scheme." << endl;
        scheme.subdivide(field[i], big);
        big.trimField( (int) (vSizeX/2), (int) (vSizeY/2), (int) (vSizeZ/2), field[i+1]);
    }

	err_log << "Filling out fine: " << vSizeX << ", " << vSizeY << ", " << vSizeZ << endl;    
    //Puts the final product into fine, provided as argument
	fine[0] = vSizeX;
	fine[1] = vSizeY;
	fine[2] = vSizeZ;
	for (i = 0; i < vSizeX * vSizeY * vSizeZ * 3; i++)
	{
		fine[3 + i] = field[num_times].giveField(i);
	}

	err_log << "Cleaning up." << endl;

	delete [] field;


	err_log << "Done." << endl;
	return 0;
}
